import logging
import time
import base64
import json
import pandas as pd
import requests
import textwrap

from typing import List, Dict, Optional, Tuple, Any

from gql import Client, gql  # type: ignore
from gql.transport.requests import RequestsHTTPTransport  # type: ignore

from transform_tools.validate import upload_configs, ERROR_RESPONSE_PREFIX
from .constants import DEFAULT_QUERY_TIMEOUT, MAX_PAGE_SIZE
from .exceptions import QueryRuntimeException
from .utils import resolve_local_commit_info
from .models import (
    MqlQueryStatus,
    MqlQueryStatusResp,
    QueryResult,
    CacheMode,
    Dimension,
    Organization,
    User,
    Materialization,
    ModelKey,
    Metric,
    MQLServer,
    MQLServerImage,
    PercentChange,
    Query,
    HealthReport,
    UserState,
    TimeGranularity,
)

TRANSFORM_PROD_API = "https://api.transformdata.io"

logger = logging.getLogger(__name__)


class BaseInterface:
    """Base implementation for our various GraphQL APIs we connect to.

    Assumes that authorization and GraphQL query pattern are the same across Backend, and MQL APIs
    """

    def __init__(self, auth_header: Dict[str, str], rest_api_url: str, verify: bool = True) -> None:  # noqa: D
        self.rest_api_url = rest_api_url
        self.auth_header = auth_header
        self.verify = verify

        self.gql_client = Client(
            transport=RequestsHTTPTransport(
                url=self.graphql_api_url, verify=self.verify, retries=2, headers=self.auth_header
            ),
            fetch_schema_from_transport=False,
        )

    def execute(  # type: ignore [misc]
        self, query: gql, verbose: bool = False, variable_values: Dict[str, Any] = None
    ) -> Dict[str, Any]:
        """Error handling for GQL Clients"""
        start_time = time.time()
        try:
            response = self.gql_client.execute(query, variable_values)
        except requests.exceptions.ConnectionError:
            raise Exception(
                textwrap.dedent(
                    """\
                    Transform could not connect to the MQL Server.

                    Check that you are currently connected to the internet or that the MQL Server is up using:

                            mql health-report
                    """
                )
            )
        except Exception as e:
            exception_response = str(e.args[0])
            if "Authentication hook unauthorized" in exception_response:
                raise Exception(
                    textwrap.dedent(
                        """\
                        Transform could not authenticate the set API Key.

                        A new API Key can be created at:

                            https://app.transformdata.io/api_keys

                        and then set using the following command:

                            mql setup -k <api-key>
                            OR instantiate with MQLClient(api_key=<api-key>)
                        """
                    )
                )
            else:
                raise
        logger.debug(f"Query for {' '.join(response.keys())} took {(time.time() - start_time):.2f}s")
        return response

    @property
    def graphql_api_url(self) -> str:  # noqa: D
        return f"{self.rest_api_url}/graphql"


class BackendInterface(BaseInterface):  # noqa: D
    def __init__(  # noqa: D
        self, auth_header: Dict[str, str], rest_api_url: str = TRANSFORM_PROD_API, verify: bool = True
    ) -> None:
        super(BackendInterface, self).__init__(auth_header, rest_api_url, verify)

    def _get_and_upload_configs(
        self, mode: str, config_dir: str, repo: str, branch: str, commit: str, is_current: Optional[bool] = None
    ) -> requests.Response:  # noqa: D
        try:
            return upload_configs(
                mode=mode,
                auth_header=self.auth_header,
                repo=repo,
                branch=branch,
                commit=commit,
                config_dir=config_dir,
                is_current=is_current,
                api_url=self.rest_api_url
            )
        except Exception as e:
            error_content = str(e).split(ERROR_RESPONSE_PREFIX)
            error_msg = str(e)
            if len(error_content) == 2:
                error_response = json.loads(error_content[1])["error"]
                error_msg = f"🙈 {error_response['error_type']}: {error_response['message']}"
            raise Exception(error_msg) from None

    def validate_configs(self, config_dir: str) -> Tuple[str, str, str]:  # noqa: D
        """Finds and validates yaml configs in the provided directory"""
        repo, branch, commit = resolve_local_commit_info(config_dir)
        self._get_and_upload_configs("validate", config_dir, repo, branch, commit)

        return repo, branch, commit

    def commit_configs(self, config_dir: str, is_current: Optional[bool] = None) -> ModelKey:
        """Finds and validates yaml configs in the provided directory"""
        repo, branch, commit = resolve_local_commit_info(config_dir)
        r = self._get_and_upload_configs("commit", config_dir, repo, branch, commit, is_current)

        return ModelKey.from_snake_dict(r.json()["model"])

    def get_my_org(self) -> Organization:  # noqa: D
        """Queries for "my org" (the current user's org)"""
        query = gql(
            """
            query MyOrgQuery {
                myOrganization {
                    id
                    name
                    createdAt
                    primaryConfigRepo
                    primaryConfigBranch
                }
            }
            """
        )

        return Organization.from_gql_dict(self.execute(query)["myOrganization"])

    def get_me(self) -> User:  # noqa: D
        """Queries for "me" (the current user)"""
        query = gql(
            """
            query MeQuery {
                myUser {
                    id
                    userName
                    email
                    mqlServerUrl
                }
            }
            """
        )

        return User.from_gql_dict(self.execute(query)["myUser"])

    def get_latest_model_key(self) -> "ModelKey":
        """Queries for latest model key in Web API GQL"""
        query = gql(
            """
            query LatestModelKeyQuery {
                myOrganization {
                    currentModel {
                        id
                        organizationId
                        gitBranch
                        gitCommit
                        gitRepo
                        createdAt
                        isCurrent
                    }
                }
            }
            """
        )
        return ModelKey.from_gql_dict(self.execute(query)["myOrganization"]["currentModel"][0])

    def get_model_key(self, model_id: int) -> "ModelKey":
        """Queries for a given model id"""
        query = gql(
            """
            query ModelKeyQuery($modelId: ID) {
                myOrganization {
                    models(id: $modelId){
                        id
                        organizationId
                        gitBranch
                        gitCommit
                        gitRepo
                        createdAt
                        isCurrent
                    }
                }
            }
            """
        )
        return ModelKey.from_gql_dict(
            self.execute(query, variable_values={"modelId": model_id})["myOrganization"]["models"][0]
        )

    def get_org_mql_servers(self) -> List[MQLServer]:  # noqa: D
        query = gql(
            """
            query MqlServersQuery {
                myUser {
                    organization {
                        mqlServers {
                            id
                            name
                            url
                            isOrgDefault
                        }
                    }
                }
            }
            """
        )

        return [MQLServer.from_gql_dict(s) for s in self.execute(query)["myUser"]["organization"]["mqlServers"]]

    def get_user_state(self) -> UserState:
        """Fetch all the core models for the User/Org/current model in one simple query"""
        query = gql(
            """
            query MeQuery {
                myUser {
                    id
                    userName
                    email
                    mqlServerUrl
                    organization {
                        id
                        name
                        createdAt
                        primaryConfigRepo
                        primaryConfigBranch
                        currentModel {
                            id
                            organizationId
                            gitBranch
                            gitCommit
                            gitRepo
                            createdAt
                            isCurrent
                        }
                    }
                }
            }
            """
        )

        resp = self.execute(query)
        user = User.from_gql_dict(resp["myUser"])
        org = Organization.from_gql_dict(resp["myUser"]["organization"])

        model: Optional[ModelKey] = None
        if len(resp["myUser"]["organization"]["currentModel"]):
            model = ModelKey.from_gql_dict(resp["myUser"]["organization"]["currentModel"][0])

        return UserState(user=user, organization=org, current_model=model)

    def get_latest_mql_server_image(self) -> MQLServerImage:
        """Fetch the latest mql server image."""
        query = gql(
            """
            query MqlLatestServerQuery {
                latestMqlServer {
                    serviceName
                    versionHash
                    downloadUrl
                }
            }
            """
        )
        return MQLServerImage.from_gql_dict(self.execute(query)["latestMqlServer"])

    def get_metrics_metadata(self) -> List[Dict[str, Any]]:  # type: ignore[misc]
        """Fetch all org metrics' metadata."""
        query = gql(
            """
            query MetricsMetadataQuery ($pageSize: Int!){
                myOrganization {
                    metrics (pageSize: $pageSize){
                        name
                        description
                        displayName
                        tier
                        userOwners{
                            user {
                                userName
                                isOrgAdmin
                                email
                            }
                        }
                        teamOwners{
                            team {
                                name
                                memberUsers {
                                    userName
                                    isOrgAdmin
                                    email
                                }
                            }
                        }
                    }
                }
            }
            """
        )
        return self.execute(query, variable_values={"pageSize": MAX_PAGE_SIZE})["myOrganization"]["metrics"]

    def get_metric_metadata_by_name(self, metric_name: str) -> Dict[str, Any]:  # type: ignore[misc]
        """Fetch a single org metric metadata."""
        query = gql(
            """
            query MetricMetadataQuery ($metricName: String!) {
                myOrganization {
                    metric (name: $metricName){
                        name
                        description
                        displayName
                        tier
                        userOwners{
                            user {
                                userName
                                isOrgAdmin
                                email
                            }
                        }
                        teamOwners{
                            team {
                                name
                                memberUsers {
                                    userName
                                    isOrgAdmin
                                    email
                                }
                            }
                        }
                    }
                }
            }
            """
        )
        return self.execute(query, variable_values={"metricName": metric_name})["myOrganization"]["metric"]


class MQLInterface(BaseInterface):  # noqa: D
    def __init__(self, auth_header: Dict[str, str], query_server_url: str) -> None:  # noqa: D
        super(MQLInterface, self).__init__(auth_header, query_server_url)

    def __common_query_input_params(  # type: ignore [misc]
        self,
        metrics: List[str],
        dimensions: List[str],
        model_key: Optional[ModelKey] = None,
        where_constraint: Optional[str] = None,
        time_constraint: Optional[str] = None,
        time_granularity: Optional[TimeGranularity] = None,
        time_comparison: Optional[PercentChange] = None,
        order: Optional[List[str]] = None,
        limit: Optional[str] = None,
        cache_mode: Optional[CacheMode] = None,
        as_table: Optional[str] = None,
        time_series: bool = True,
        result_format: Optional[str] = None,
        allow_dynamic_cache: bool = True,
    ) -> Dict[str, Any]:
        return {
            "modelKey": {
                "organization": model_key.organization_id,
                "repo": model_key.repository,
                "branch": model_key.branch,
                "commit": model_key.commit,
            }
            if model_key
            else None,
            "metrics": metrics,
            "groupBy": dimensions,
            "whereConstraint": where_constraint,
            "timeConstraint": time_constraint,
            "timeGranularity": time_granularity.name if time_granularity else None,
            "pctChange": time_comparison.name if time_comparison else None,
            "order": order,
            "limit": limit,
            "cacheMode": (CacheMode.default() if cache_mode is None else cache_mode).name,
            "asTable": as_table,
            "timeSeries": time_series,
            "resultFormat": result_format,
            "allowDynamicCache": allow_dynamic_cache,
        }

    def drop_materialization(  # noqa: D
        self,
        materialization_name: str,
        start_time: Optional[str],
        end_time: Optional[str],
        model_key: Optional[ModelKey] = None,
    ) -> str:
        """Implements drop materialization mutation in MQL/GQL"""
        query = gql(
            """
            mutation CreateMqlDropMaterializationMutation(
                $materializationName: String!
                $startTime: String
                $endTime: String
                $modelKey: ModelKeyInput
            ) {
                createMqlDropMaterialization(
                    input: {
                        modelKey: $modelKey,
                        materializationName: $materializationName,
                        startTime: $startTime,
                        endTime: $endTime,
                    }
                ) {
                    id
                }
            }
            """
        )

        return self.execute(
            query,
            variable_values={
                "modelKey": {
                    "organization": model_key.organization_id,
                    "repo": model_key.repository,
                    "branch": model_key.branch,
                    "commit": model_key.commit,
                }
                if model_key
                else None,
                "materializationName": materialization_name,
                "startTime": start_time,
                "endTime": end_time,
            },
        )["createMqlDropMaterialization"]["id"]

    def create_materialization(  # noqa: D
        self,
        materialization_name: str,
        start_time: Optional[str],
        end_time: Optional[str],
        model_key: Optional[ModelKey] = None,
        output_table: Optional[str] = None,
        force: bool = False,
    ) -> str:
        """Implements Materialize (new) mutation in MQL/GQL"""
        # Note: we omit the outputTable var if arg is not passed to maintain backwards compatibility with servers
        # that don't recognize it (ie if they're not on 0e7448e8b6a75484f806c32c7f74e3aafbf11b7d or later)
        query = gql(
            """
            mutation CreateMqlMaterializationNewMutation(
                $materializationName: String!
                $startTime: String
                $endTime: String
                $modelKey: ModelKeyInput
                $outputTable: String
                $force: Boolean
            ) {
                createMqlMaterializationNew(
                    input: {
                        modelKey: $modelKey,
                        materializationName: $materializationName,
                        startTime: $startTime,
                        endTime: $endTime,
                        outputTable: $outputTable
                        force: $force
                    }
                ) {
                    id
                }
            }
            """
        )

        return self.execute(
            query,
            variable_values={
                "modelKey": {
                    "organization": model_key.organization_id,
                    "repo": model_key.repository,
                    "branch": model_key.branch,
                    "commit": model_key.commit,
                }
                if model_key
                else None,
                "materializationName": materialization_name,
                "startTime": start_time,
                "endTime": end_time,
                "outputTable": output_table,
                "force": force,
            },
        )["createMqlMaterializationNew"]["id"]

    def get_materializations(self, model_key: Optional[ModelKey] = None) -> List[Materialization]:  # noqa: D
        query = gql(
            """
            query MaterializationsList($modelKey: ModelKeyInput) {
                materializations(modelKey: $modelKey) {
                    name
                    metrics
                    dimensions
                    destinationTable
                }
            }
            """
        )

        return [
            Materialization.from_gql_dict(m)
            for m in self.execute(
                query,
                variable_values={
                    "modelKey": {
                        "organization": model_key.organization_id,
                        "repo": model_key.repository,
                        "branch": model_key.branch,
                        "commit": model_key.commit,
                    }
                    if model_key
                    else None
                },
            )["materializations"]
        ]

    def get_dimension_values(
        self,
        metric_name: str,
        dimension_name: str,
        allow_dynamic_cache: bool = True,
        model_key: Optional[ModelKey] = None,
        page_number: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[str]:
        """Retrieve all dimension values of a given [metric,dimension].

        NOTE: dynamic caching is disabled (ie., only allowed to retrieve dimension values through materialization)
        """
        query = gql(
            """
            query DimensionValueQuery($dimensionName: String!, $metricName: String!, $allowDynamicCache: Boolean, $modelKey: ModelKeyInput, $pageNumber: Int, $pageSize: Int) {
                metricByName(modelKey: $modelKey, name: $metricName) {
                    dimensionValues(dimensionName: $dimensionName, allowDynamicCache: $allowDynamicCache, pageNumber: $pageNumber, pageSize: $pageSize)
                }
            }
            """
        )

        return self.execute(
            query,
            variable_values={
                "modelKey": {
                    "organization": model_key.organization_id,
                    "repo": model_key.repository,
                    "branch": model_key.branch,
                    "commit": model_key.commit,
                }
                if model_key
                else None,
                "pageNumber": page_number,
                "pageSize": page_size,
                "dimensionName": dimension_name,
                "metricName": metric_name,
                "allowDynamicCache": allow_dynamic_cache,
            },
        )["metricByName"]["dimensionValues"]

    def get_metrics(self, model_key: Optional[ModelKey] = None) -> List[Metric]:  # noqa: D
        query = gql(
            """
            query MetricList($modelKey: ModelKeyInput) {
                metrics(modelKey: $modelKey) {
                    name
                    measures
                    type
                    typeParams {
                        numerator
                        denominator
                        expr
                        window
                    }
                    constraint
                    dimensionObjects {
                        name
                        identifierName
                        identifierNames
                        type
                        isPrimaryTime
                        timeGranularity
                        cardinality
                    }
                }
            }
            """
        )
        return [
            Metric.from_gql_dict(m)
            for m in self.execute(
                query,
                variable_values={
                    "modelKey": {
                        "organization": model_key.organization_id,
                        "repo": model_key.repository,
                        "branch": model_key.branch,
                        "commit": model_key.commit,
                    }
                    if model_key
                    else None
                },
            )["metrics"]
        ]

    def get_metric_by_name(self, metric_name: str, model_key: Optional[ModelKey] = None) -> Metric:  # noqa: D
        query = gql(
            """
            query MetricQuery($modelKey: ModelKeyInput, $metricName: String!) {
                metricByName(modelKey: $modelKey, name: $metricName) {
                    name
                    measures
                    type
                    typeParams {
                        numerator
                        denominator
                        expr
                        window
                    }
                    constraint
                    dimensionObjects {
                        name
                        identifierName
                        identifierNames
                        type
                        isPrimaryTime
                        timeGranularity
                        cardinality
                    }
                }
            }
            """
        )
        return Metric.from_gql_dict(
            self.execute(
                query,
                variable_values={
                    "modelKey": {
                        "organization": model_key.organization_id,
                        "repo": model_key.repository,
                        "branch": model_key.branch,
                        "commit": model_key.commit,
                    }
                    if model_key
                    else None,
                    "metricName": metric_name,
                },
            )["metricByName"]
        )

    def get_dimensions(self, model_key: Optional[ModelKey] = None) -> List[Dimension]:
        """Get a list of all unique Dimension objects."""
        query = gql(
            """
            query DimensionsQuery($modelKey: ModelKeyInput) {
                dimensions(modelKey: $modelKey) {
                    name
                    identifierName
                    identifierNames
                    type
                    isPrimaryTime
                    timeGranularity
                    cardinality
                }
            }
            """
        )
        return [
            Dimension.from_gql_dict(dim)
            for dim in self.execute(
                query,
                variable_values={
                    "modelKey": {
                        "organization": model_key.organization_id,
                        "repo": model_key.repository,
                        "branch": model_key.branch,
                        "commit": model_key.commit,
                    }
                    if model_key
                    else None
                },
            )["dimensions"]
        ]

    def create_query(  # noqa: D
        self,
        metrics: List[str],
        dimensions: List[str],
        model_key: Optional[ModelKey] = None,
        where_constraint_str: Optional[str] = None,
        time_constraint: Optional[str] = None,
        time_granularity: Optional[TimeGranularity] = None,
        time_comparison: Optional[PercentChange] = None,
        order: Optional[List[str]] = None,
        limit: Optional[str] = None,
        cache_mode: Optional[CacheMode] = None,
        as_table: Optional[str] = None,
        allow_dynamic_cache: bool = True,
    ) -> str:
        """Implements CreateMqlQuery mutation in MQL/GQL"""

        # TODO: Find a fix to being able to show cli query graph on server logs
        time_series = False
        result_format = None
        query = gql(
            """
            mutation CreateMqlQueryMutation(
                $modelKey: ModelKeyInput,
                $metrics: [String!],
                $groupBy: [String!],
                $whereConstraint: String,
                $timeConstraint: String,
                $timeGranularity: TimeGranularity,
                $order: [String!],
                $limit: LimitInput,
                $cacheMode: CacheMode,
                $asTable: String,
                $timeSeries: Boolean,
                $resultFormat: ResultFormat,
                $pctChange: PercentChange,
                $allowDynamicCache: Boolean,
            ) {
                createMqlQuery(
                    input: {
                        modelKey: $modelKey,
                        metrics: $metrics,
                        groupBy: $groupBy,
                        whereConstraint: $whereConstraint,
                        timeConstraint: $timeConstraint,
                        timeGranularity: $timeGranularity,
                        order: $order,
                        limit: $limit,
                        cacheMode: $cacheMode,
                        asTable: $asTable,
                        addTimeSeries: $timeSeries,
                        resultFormat: $resultFormat,
                        pctChange: $pctChange,
                        allowDynamicCache: $allowDynamicCache
                    }
                ) {
                    id
                }
            }
            """
        )

        params = self.__common_query_input_params(
            model_key=model_key,
            metrics=metrics,
            dimensions=dimensions,
            where_constraint=where_constraint_str,
            time_constraint=time_constraint,
            time_granularity=time_granularity,
            time_comparison=time_comparison,
            order=order,
            limit=limit,
            cache_mode=cache_mode,
            as_table=as_table,
            time_series=time_series,
            result_format=result_format,
            allow_dynamic_cache=allow_dynamic_cache,
        )

        return self.execute(query, variable_values=params)["createMqlQuery"]["id"]

    def drop_cache(self) -> bool:  # noqa: D
        """Drops the entire MQL cache. Only do this is the cache is somehow corrupt."""
        query = gql(
            """
            mutation {
                dropCache(confirm: "yes") {
                    success
                }
            }
            """
        )

        return bool(self.execute(query)["dropCache"]["success"])

    def ping(self) -> requests.Response:  # noqa: D
        """Calls the MQL REST API for Health"""
        return requests.get(f"{self.rest_api_url}/health", headers=self.auth_header, timeout=10)

    def get_version(self) -> str:  # noqa: D
        query = gql(
            """
            query GetVersion {
                version
            }
            """
        )
        return self.execute(query)["version"]

    def get_health_report(self) -> Tuple[str, List[HealthReport]]:  # noqa: D
        """Calls the MQL Server GQL API for Health"""
        query = gql(
            """
            query GetHealthReport {
                version
                healthReport {
                    name
                    status
                    errorMessage
                }
            }
            """
        )

        query_result = self.execute(query)

        return query_result["version"], [HealthReport.from_gql_dict(h) for h in query_result["healthReport"]]

    def _get_materialization_result(self, query_id: str) -> Tuple[str, str]:
        """Retrieves query status given a query_id"""
        query = gql(
            """
            query GetMqlQueryResultsStatus($queryId: ID!) {
                mqlQuery(id: $queryId) {
                    resultTableSchema,
                    resultTableName
                }
            }
            """
        )

        query_result = self.execute(query, variable_values={"queryId": query_id})["mqlQuery"]
        return query_result["resultTableSchema"], query_result["resultTableName"]

    def explain_query_sql(self, query_id: str) -> str:
        """Retrieves the SQL generated by the MQL server for this query"""
        query = gql(
            """
            query SourceQuery($queryId: ID!) {
                sourceQuery(id: $queryId)
            }
            """
        )

        return self.execute(query, variable_values={"queryId": query_id})["sourceQuery"]

    def get_query_status(self, query_id: str) -> MqlQueryStatusResp:
        """Retrieves query status given a query_id"""
        query = gql(
            """
            query GetMqlQueryResultsStatus($queryId: ID!) {
                mqlQuery(id: $queryId) {
                    status
                    error
                    sql
                }
            }
            """
        )

        q = self.execute(query, variable_values={"queryId": query_id})["mqlQuery"]
        return MqlQueryStatusResp(
            query_id=query_id,
            status=MqlQueryStatus[q["status"]],
            error=q["error"] if q["error"] else None,
        )

    def get_queries(self, active_only: bool, limit: Optional[int]) -> List[Query]:
        """Retrieves query status given a query_id"""
        query = gql(
            """
            query GetMqlQueries($activeOnly: Boolean, $limit: Int) {
                queries(activeOnly: $activeOnly, limit:$limit) {
                    id,
                    modelKey {
                        branch,
                        commit,
                    },
                    metrics,
                    dimensions,
                    status,
                }
            }
            """
        )
        return [
            Query.from_gql_dict(q)
            for q in self.execute(query, variable_values={"activeOnly": active_only, "limit": limit})["queries"]
        ]

    def get_logs_by_line(self, query_id: str, from_line: int, max_lines: int = 0) -> Tuple[str, int]:
        """Retrieves logs by log-file line number"""
        query = gql(
            """
            query GetLogLines($queryId: ID!, $fromLine: Int, $maxLines: Int) {
                mqlQuery(id: $queryId) {
                    logsByLine(fromLine: $fromLine, maxLines: $maxLines),
                }
            }
            """
        )
        variable_values = {"queryId": query_id, "fromLine": from_line, "maxLines": max_lines}
        logs = self.execute(query, variable_values=variable_values)["mqlQuery"]["logsByLine"]
        return logs, len(logs.split("\n"))

    def get_query_page_as_df(self, query_id: str, cursor: Optional[int] = None) -> QueryResult:
        """Retrieves a single page of a query's results and converts to a DataFrame"""
        cursor = cursor or 0
        query = gql(
            """
            query GetMqlQueryResultsTabular($queryId: ID!, $cursor: Int) {
                mqlQuery(id: $queryId) {
                    resultTabular(orient: TABLE, cursor: $cursor) {
                        nextCursor
                        data
                    }
                }
            }
            """
        )

        query_result = self.execute(query, variable_values={"queryId": query_id, "cursor": cursor})["mqlQuery"]
        tabular = query_result["resultTabular"]
        df = pd.read_json(
            base64.b64decode(tabular["data"].encode()),
            orient="table",
        )
        return QueryResult(
            df=df,
            cursor=tabular["nextCursor"],
        )

    def poll_for_query_completion(self, query_id: str, timeout: Optional[int] = None) -> MqlQueryStatusResp:
        """Poll for query completion, displaying some progress indication."""
        timeout = timeout if timeout is not None else DEFAULT_QUERY_TIMEOUT
        start = time.time()

        # Poll until the query is complete or we hit the timeout (timeout of 0 indicate no timeout)
        poll_interval = 0.1
        while timeout == 0 or time.time() < start + timeout:
            query_result = self.get_query_status(query_id)
            if query_result.is_complete:
                return query_result
            time.sleep(poll_interval)

            # poll interval starts very low and gets longer until we hit a maximum of 5s
            # allows us to resolve fast queries quickly without hammering mql for long queries
            poll_interval = min(5, poll_interval * 1.5)

        raise QueryRuntimeException(
            query_id,
            f"Timeout reached waiting for query {query_id} to complete after {timeout} seconds. Please see --timeout option to override.",
        )

    def get_materialization_result(self, query_id: str, timeout: Optional[int] = None) -> Tuple[str, str]:
        """Retrieves materialized table resulting from successful materialization query"""
        resp = self.poll_for_query_completion(query_id, timeout=timeout)
        if not resp.is_successful:
            raise QueryRuntimeException.from_query_response(resp)

        return self._get_materialization_result(query_id)

    def get_query_dataframe(self, query_id: str, timeout: Optional[int] = None) -> pd.DataFrame:
        """Retrieves query results in Pandas DataFrame format

        Includes logic for:
        1. Polling for completion
        2. Paging through results
        3. Converting JSON to a Pandas DataFrame
        """
        resp = self.poll_for_query_completion(query_id, timeout=timeout)
        if not resp.is_successful:
            raise QueryRuntimeException.from_query_response(resp)

        cursor: Optional[int] = 0
        df = pd.DataFrame()

        while cursor is not None:
            query_result = self.get_query_page_as_df(query_id, cursor)
            cursor = query_result.cursor
            df = df.append(query_result.df, ignore_index=True)

        return df



import pandas as pd
import arrow

from typing import List, Optional, Any, Dict
from enum import Enum
from dataclasses import dataclass, field

INF = "inf"
NEGATIVE_ONE = "-1"
DEFAULT_LIMIT = "100"


@dataclass(frozen=True)
class ModelKey:
    """A model key. Represents a given point in time for the Transform configs."""

    id: int
    organization_id: int
    repository: str
    branch: str
    commit: str
    created_at_ts: str
    is_current: bool

    @property
    def created_at(self) -> str:
        """Converts a system timestamp to a human readable string, ex: 'an hour ago'"""
        return arrow.get(self.created_at_ts).humanize()

    @classmethod
    def from_gql_dict(cls, _input: Dict[str, str]) -> "ModelKey":  # noqa: D
        return cls(
            id=int(_input["id"]),
            organization_id=int(_input["organizationId"]),
            repository=_input["gitRepo"],
            branch=_input["gitBranch"],
            commit=_input["gitCommit"],
            created_at_ts=_input["createdAt"],
            is_current=bool(_input["isCurrent"]),
        )

    @classmethod
    def from_snake_dict(cls, _input: Dict[str, str]) -> "ModelKey":  # noqa: D
        return cls(
            id=int(_input["id"]),
            organization_id=int(_input["organization_id"]),
            repository=_input["repository"],
            branch=_input["branch"],
            commit=_input["commit"],
            created_at_ts=_input["created_at"],
            is_current=bool(_input["is_current"]),
        )


@dataclass(frozen=True)
class Organization:
    """Lightweight object to represent a Transform Organization."""

    id: int
    name: str
    primary_config_repo: str
    primary_config_branch: str

    @classmethod
    def from_gql_dict(cls, _input: Dict[str, str]) -> "Organization":  # noqa: D
        return cls(
            id=int(_input["id"]),
            name=_input["name"],
            primary_config_repo=_input["primaryConfigRepo"],
            primary_config_branch=_input["primaryConfigBranch"],
        )


@dataclass(frozen=True)
class User:
    """Lightweight object to represent a Transform User."""

    id: int
    user_name: str
    email: str
    mql_server_url: str

    @classmethod
    def from_gql_dict(cls, _input: Dict[str, str]) -> "User":  # noqa: D
        return cls(
            id=int(_input["id"]),
            user_name=_input["userName"],
            email=_input["email"],
            mql_server_url=_input["mqlServerUrl"],
        )


@dataclass(frozen=True)
class UserState:
    """Aggregation of the core models necessary for most operations"""

    user: User
    organization: Organization
    current_model: Optional[ModelKey]


@dataclass(frozen=True)
class Materialization:
    """Object to represent a Metric."""

    name: str
    metrics: List[str]
    dimensions: List[str]
    destination_table: Optional[str]

    @classmethod
    def from_gql_dict(cls, _input: Dict[str, Any]) -> "Materialization":  # type: ignore [misc] # noqa: D
        destination_table = None
        if "destination_table" in _input:
            destination_table = _input["destination_table"]
        return cls(
            name=_input["name"],
            metrics=_input["metrics"],
            dimensions=_input["dimensions"],
            destination_table=destination_table,
        )


@dataclass(frozen=True)
class UserOwner:
    """Object to represent a user owner."""

    user_name: str
    is_org_admin: bool
    email: str

    @classmethod
    def from_gql_dict(cls, _input: Dict[str, Any]) -> "UserOwner":  # type: ignore [misc] # noqa: D
        return cls(
            user_name=_input["userName"],
            is_org_admin=_input["isOrgAdmin"],
            email=_input["email"],
        )


@dataclass(frozen=True)
class TeamOwner:
    """Object to represent a team owner."""

    name: str
    users: List[UserOwner]

    @classmethod
    def from_gql_dict(cls, _input: Dict[str, Any]) -> "TeamOwner":  # type: ignore [misc] # noqa: D
        return cls(
            name=_input["name"],
            users=[UserOwner.from_gql_dict(u) for u in _input["memberUsers"]],
        )


class MetricType(Enum):
    """Metric type Enum."""

    MEASURE_PROXY = "measure_proxy"
    RATIO = "ratio"
    EXPR = "expr"
    CUMULATIVE = "cumulative"


@dataclass
class MetricTypeParams:
    """Object to represent the type params of a Metric."""

    numerator: Optional[str]
    denominator: Optional[str]
    expr: Optional[str]
    window: Optional[str]

    @classmethod
    def from_gql_dict(cls, _input: Dict[str, Any]) -> "MetricTypeParams":  # type: ignore [misc] # noqa: D
        return cls(
            numerator=_input["numerator"],
            denominator=_input["denominator"],
            expr=_input["expr"],
            window=_input["window"],
        )


@dataclass
class Metric:
    """Object to represent a Metric."""

    name: str
    measures: List[str]
    metric_type: MetricType
    type_params: MetricTypeParams
    constraint: Optional[str]
    dimensions: List["Dimension"]
    team_owners: Optional[List[TeamOwner]] = None
    user_owners: Optional[List[UserOwner]] = None
    display_name: Optional[str] = None
    description: Optional[str] = None
    tier: Optional[int] = None

    @classmethod
    def from_gql_dict(cls, _input: Dict[str, Any]) -> "Metric":  # type: ignore [misc] # noqa: D
        return cls(
            name=_input["name"],
            measures=_input["measures"],
            metric_type=MetricType(_input["type"].lower()),
            type_params=MetricTypeParams.from_gql_dict(_input["typeParams"]),
            constraint=_input["constraint"],
            dimensions=list({d["name"]: Dimension.from_gql_dict(d) for d in _input["dimensionObjects"]}.values()),
        )

    def input_metadata(self, metadata: Dict[str, Any]) -> None:  # type: ignore[misc]
        """Input additional backend metadata."""
        self.description = metadata["description"]
        self.display_name = metadata["displayName"]
        self.tier = metadata["tier"]
        self.team_owners = [TeamOwner.from_gql_dict(t["team"]) for t in metadata["teamOwners"]]
        self.user_owners = [UserOwner.from_gql_dict(u["user"]) for u in metadata["userOwners"]]


class MqlQueryStatus(Enum):
    """The status of queries submitted for execution in the query manager."""

    # Created and waiting to run.
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    SUCCESSFUL = "SUCCESSFUL"
    # Handled exception prevented a query from being successful.
    FAILED = "FAILED"
    # Unhandled exception prevented a query from being successful.
    UNHANDLED_EXCEPTION = "UNHANDLED_EXCEPTION"
    # The given query ID is not known to exist - possibly expired.
    UNKNOWN = "UNKNOWN"
    TIMEOUT = "TIMEOUT"


@dataclass(frozen=True)
class MqlQueryStatusResp:  # noqa: D
    query_id: str
    status: MqlQueryStatus
    error: Optional[str]

    @property
    def is_complete(self) -> bool:
        """These statuses indicate the Query has completed execution"""
        return self.status in [
            None,
            MqlQueryStatus.SUCCESSFUL,
            MqlQueryStatus.FAILED,
            MqlQueryStatus.UNHANDLED_EXCEPTION,
        ]

    @property
    def is_successful(self) -> bool:
        """Indicate whether the Query has completed successfully"""
        return self.status == MqlQueryStatus.SUCCESSFUL

    @property
    def is_failed(self) -> bool:
        """Indicate whether the Query has completed successfully"""
        return self.status in [MqlQueryStatus.FAILED, MqlQueryStatus.UNHANDLED_EXCEPTION]


@dataclass(frozen=True)
class QueryResult:
    """Object for query results with a cursor"""

    df: pd.DataFrame
    cursor: Optional[int] = None


class CacheMode(Enum):
    """Different CacheMode options."""

    READ = "r"  # use cached datasets but do not write new datasets to cache
    READWRITE = "rw"  # use cached datasets and write new datasets to cache
    WRITE = "w"  # ignore cached datatsets and write new datasets to cache
    IGNORE = "i"  # ignore cache completely, only use configuration files

    @classmethod
    def from_string(cls, string_value: str) -> "CacheMode":  # noqa: D
        if string_value == "wr":  # convenience rewire
            string_value = "rw"
        return cls(string_value)

    @classmethod
    def default(cls) -> "CacheMode":  # noqa: D
        return CacheMode.READWRITE


class TimeGranularity(Enum):
    """For time dimensions, the smallest possible difference between to time values.

    Needed for calculating adjacency when merging 2 different time ranges.
    """

    # Names are used in parameters to DATE_TRUC, so don't change them.
    # Values are used to convert user supplied string to enums.
    DAY = "day"
    WEEK = "week"
    MONTH = "month"
    QUARTER = "quarter"
    YEAR = "year"


@dataclass(frozen=True)
class MQLServer:
    """Object to represent a Org MQL Server."""

    id: int
    name: str
    url: str
    is_org_default: bool

    @classmethod
    def from_gql_dict(cls, _input: Dict[str, str]) -> "MQLServer":  # noqa: D
        return cls(
            id=int(_input["id"]),
            name=_input["name"],
            url=_input["url"],
            is_org_default=bool(_input["isOrgDefault"]),
        )


class HealthReportStatus(Enum):
    """Statuses for Health Reports."""

    SUCCESS = "SUCCESS"  # MQL Server responded as healthy
    FAIL = "FAIL"  # MQL Server responded as unhealthy


@dataclass(frozen=True)
class HealthReport:
    """Object to represent a MQL Server Health Report."""

    name: str
    status: str
    error_message: str

    @classmethod
    def from_gql_dict(cls, _input: Dict[str, str]) -> "HealthReport":  # noqa: D
        return cls(
            name=_input["name"],
            status=_input["status"],
            error_message=_input["errorMessage"],
        )


@dataclass(frozen=True)
class ServerHealthReport:
    """Object to represent the health of a data warehouse."""

    name: str
    status: str
    url: str
    version: str = ""
    error_message: str = ""
    servers: List[HealthReport] = field(default_factory=list)


@dataclass(frozen=True)
class MqlMaterializeResp:
    """Response object received from MQL from a successful Materialize request."""

    schema: Optional[str]
    table: Optional[str]
    query_id: str

    @property
    def fully_qualified_name(self) -> str:  # noqa: D
        return f"{self.schema}.{self.table}"


@dataclass(frozen=True)
class MQLServerImage:
    """Object to represent a MQL Server Image."""

    service_name: str
    version_hash: str
    download_url: str

    @classmethod
    def from_gql_dict(cls, _input: Dict[str, str]) -> "MQLServerImage":  # noqa: D
        return cls(
            service_name=_input["serviceName"],
            version_hash=_input["versionHash"],
            download_url=_input["downloadUrl"],
        )


class PercentChange(Enum):
    """Used to calculate the percent change between the given granularity."""

    DOD = "dod"  # day-over-day
    WOW = "wow"  # week-over-week
    MOM = "mom"  # month-over-month
    QOQ = "qoq"  # quarter-over-quarter
    YOY = "yoy"  # year-over-year


class DimensionType(Enum):
    """Dimension type Enums."""

    CATEGORICAL = "categorical"
    TIME = "time"


@dataclass(frozen=True)
class Dimension:
    """Object to represent a Dimension object."""

    name: str
    identifier_name: str
    identifier_names: List[str]
    dimension_type: DimensionType
    is_primary_time: bool
    time_granularity: Optional[TimeGranularity]
    cardinality: Optional[int]

    @classmethod
    def from_gql_dict(cls, _input: Dict[str, Any]) -> "Dimension":  # type: ignore[misc] # noqa: D
        return cls(
            name=_input["name"],
            identifier_name=_input["identifierName"],
            identifier_names=_input["identifierNames"],
            dimension_type=DimensionType(_input["type"].lower()),
            is_primary_time=_input["isPrimaryTime"],
            time_granularity=TimeGranularity(_input["timeGranularity"].lower()) if _input["timeGranularity"] else None,
            cardinality=_input["cardinality"],
        )


@dataclass(frozen=True)
class Query:
    """Object to represent a Query object."""

    id: int
    branch: int
    commit: str
    metrics: List[str]
    dimensions: List[str]
    status: MqlQueryStatus

    @classmethod
    def from_gql_dict(cls, _input: Dict[str, Any]) -> "Query":  # type: ignore[misc] # noqa: D
        return cls(
            id=_input["id"],
            branch=_input["modelKey"]["branch"],
            commit=_input["modelKey"]["commit"],
            metrics=_input["metrics"],
            dimensions=_input["dimensions"],
            status=MqlQueryStatus(_input["status"]),
        )



import os
import pathlib
import validators
import yaml

from dateutil import parser
from datetime import datetime
from typing import Dict, Optional
from pathlib import Path

from .constants import (
    API_KEY_CONFIG_KEY,
    MQL_OVERRIDE_CONFIG_KEY,
    BEARER_TOKEN_CONFIG_KEY,
    BEARER_TOKEN_EXPIRES_AT_CONFIG_KEY,
)
from .exceptions import AuthException, URLException
from .interfaces import MQLInterface, BackendInterface
from .models import User, Organization, ModelKey, UserState


class TransformAuth:  # noqa: D
    def __init__(  # noqa: D
        self, api_key: Optional[str] = None, mql_server_url: Optional[str] = None, override_config: bool = True
    ) -> None:
        self.__user_state: Optional[UserState] = None
        self.__mql_server_url: Optional[str] = None
        self.__mql_client: Optional[MQLInterface] = None
        self.__auth_header: Optional[Dict[str, str]] = None
        self.config = ConfigHandler(override_config)

        # Overrides
        if api_key is not None:
            self.api_key = api_key
        if mql_server_url is not None:
            self.mql_server_url_config_override = mql_server_url

    @property
    def api_key(self) -> Optional[str]:
        """Pull API key from config file by default, also supporting a fallback environment variable TRANSFORM_API_KEY."""
        config_api_key = self.config._get_config_value(API_KEY_CONFIG_KEY)
        return config_api_key if config_api_key else os.getenv("TRANSFORM_API_KEY")

    @api_key.setter
    def api_key(self, api_key: str) -> None:
        """Set the users' API key and record to the config file."""

        # Validate API Key we expect a format like tfdk-prefix-secret
        # Do some very basic validation that this fits before setting key
        if not api_key.count("-") == 2:
            raise AuthException

        self.config._set_config_value(API_KEY_CONFIG_KEY, api_key)

    @property
    def bearer_token(self) -> Optional[str]:
        """Get bearer token from config, if exists."""
        token = self.config._get_config_value(BEARER_TOKEN_CONFIG_KEY)
        if (
            token is not None
            and self.bearer_token_expires_at is not None
            and parser.parse(self.bearer_token_expires_at) <= datetime.now()
        ):
            raise AuthException("Bearer token has expired")

        return token

    @bearer_token.setter
    def bearer_token(self, bearer_token: str) -> None:  # noqa: D
        self.config._set_config_value(BEARER_TOKEN_CONFIG_KEY, bearer_token)

    @property
    def bearer_token_expires_at(self) -> Optional[str]:
        """Time that current bearer token will expire, if exists."""
        return self.config._get_config_value(BEARER_TOKEN_EXPIRES_AT_CONFIG_KEY)

    @bearer_token_expires_at.setter
    def bearer_token_expires_at(self, bearer_token_expires_at: str) -> None:  # noqa: D
        self.config._set_config_value(BEARER_TOKEN_EXPIRES_AT_CONFIG_KEY, bearer_token_expires_at)

    @property
    def auth_header(self) -> Dict[str, str]:
        """Get auth header based on creds in config file. If none, return None."""
        if not self.__auth_header:
            if self.api_key is not None:
                self.__auth_header = {"Authorization": f"X-Api-Key {self.api_key}"}
            elif self.bearer_token is not None:
                self.__auth_header = {"Authorization": f"Bearer {self.bearer_token}"}
        if self.__auth_header is None:
            raise AuthException
        return self.__auth_header

    @property
    def backend_client(self) -> BackendInterface:
        """Returns backend client"""
        return BackendInterface(auth_header=self.auth_header)

    @property
    def user_state(self) -> UserState:  # noqa: D
        if self.__user_state is None:
            self.__user_state = self.backend_client.get_user_state()
        return self.__user_state

    @property
    def org(self) -> Organization:  # noqa: D
        return self.user_state.organization

    @property
    def user(self) -> User:  # noqa: D
        return self.user_state.user

    @property
    def current_model(self) -> Optional[ModelKey]:  # noqa: D
        return self.user_state.current_model

    @property
    def mql_server_url_status(self) -> str:  # noqa: D
        return f"Your MQL server is {self.mql_server_url} " + (
            f"which is overriden via {self.config.config_file_path} from {self.user.mql_server_url}"
            if self.mql_server_url_config_override
            else ""
        )

    @property
    def mql_server_url(self) -> str:
        """The URL for the org's MQL server.

        Overrideable via passing MQL server url at instantiation process for users who need to redirect using a local proxy.
        """
        if self.mql_server_url_config_override:
            return self.mql_server_url_config_override
        if not self.__mql_server_url:
            self.__mql_server_url = self.user.mql_server_url
        return self.__mql_server_url

    @property
    def mql_server_url_config_override(self) -> Optional[str]:
        """Config override for the users' MQL server URL

        Overrideable via passing MQL server url at instantiation process for users who need to redirect using a local proxy.
        """
        return self.config._get_config_value(MQL_OVERRIDE_CONFIG_KEY)

    @mql_server_url_config_override.setter
    def mql_server_url_config_override(self, mql_server_url: str) -> None:  # noqa: D
        if not mql_server_url:
            self.config._remove_config_value(MQL_OVERRIDE_CONFIG_KEY)
        elif validators.url(mql_server_url, public=False):
            self.config._set_config_value(MQL_OVERRIDE_CONFIG_KEY, mql_server_url)
        else:
            raise URLException(mql_server_url)

    @property
    def mql_client(self) -> MQLInterface:  # noqa: D
        if not self.__mql_client:
            self.__mql_client = MQLInterface(query_server_url=self.mql_server_url, auth_header=self.auth_header)
        return self.__mql_client


class ConfigHandler:
    """Class for handling interactions with config.yml"""

    def __init__(self, override_config: bool = True) -> None:  # noqa: D
        # Check if the proper directory exists, if not create it
        if not os.path.exists(self.config_dir):
            cofig_dir = pathlib.Path(self.config_dir)
            cofig_dir.mkdir(parents=True)
        self.override_config = override_config

    @property
    def config_dir(self) -> str:
        """Retrieve Transform config_dir from $TFD_CONFIG_DIR, default config dir is ~/.transform"""
        config_dir_env = os.getenv("TFD_CONFIG_DIR")

        return config_dir_env if config_dir_env else f"{str(Path.home())}/.transform"

    @property
    def config_file_path(self) -> str:
        """Config file can be found at <config_dir>/config.yml"""
        return os.path.join(self.config_dir, "config.yml")

    def _get_config_value(self, key: str) -> Optional[str]:
        try:
            config_file = open(self.config_file_path, "r")
        except FileNotFoundError:
            return None

        config = yaml.load(config_file, Loader=yaml.Loader)
        if config and key in config and config[key]:
            return config[key]

        return None

    def _set_config_value(self, key: str, value: str) -> None:
        # If the caller passed a falsey value, assume they'd like to remove the key
        if not self.override_config:
            return  # Don't override config if flagged
        if not value:
            return self._remove_config_value(key)

        configs: Dict[str, str] = {}
        if os.path.exists(self.config_file_path):
            with open(self.config_file_path) as f:
                configs = yaml.load(f, Loader=yaml.SafeLoader) or {}

        configs[key] = value

        with open(self.config_file_path, "w") as f:
            yaml.dump(configs, f)

    def _remove_config_value(self, key: str) -> None:
        if not self.override_config:
            return  # Don't override config if flagged
        configs: Dict[str, str] = {}
        if os.path.exists(self.config_file_path):
            with open(self.config_file_path) as f:
                configs = yaml.load(f, Loader=yaml.SafeLoader) or {}

        del configs[key]

        with open(self.config_file_path, "w") as f:
            yaml.dump(configs, f)



DEFAULT_QUERY_TIMEOUT = 180
MAX_PAGE_SIZE = 1000  # arbitrary large number to return all results

API_KEY_CONFIG_KEY = "api_key"
BEARER_TOKEN_CONFIG_KEY = "bearer_token"
BEARER_TOKEN_EXPIRES_AT_CONFIG_KEY = "bearer_token_expires_at"
MQL_OVERRIDE_CONFIG_KEY = "mql_server_url_override"
PINNED_MODEL_ID_CONFIG_KEY = "pinned_model_id"



import os
import ast
import logging
import click

from datetime import datetime
from functools import wraps
from typing import Any, Callable, List, Optional

from ..constants import DEFAULT_QUERY_TIMEOUT
from ..models import CacheMode, PercentChange, INF, NEGATIVE_ONE, DEFAULT_LIMIT, TimeGranularity


class PythonLiteralOption(click.Option):
    """Click class to decode a string literal into a Python object."""

    def type_cast_value(self, ctx, value):  # type: ignore
        """Cast to Python obj, throws exception if we are unable to cast"""

        # For non-required inputs we receive a None here, and should return it
        if value is None:
            return None
        try:
            return ast.literal_eval(value)
        except Exception:
            raise click.BadParameter(value)


def query_observability_options(function: Callable) -> Callable:
    """Options for display mql server log"""
    function = click.option(
        "--debug",
        required=False,
        is_flag=True,
        default=False,
        help="Enable showing query log in the terminal for debugging purpose",
    )(function)

    function = click.option(
        "--web",
        required=False,
        is_flag=True,
        default=False,
        help="Open the MQL server logs on the web",
    )(function)

    return function


def async_options(function: Callable) -> Callable:
    """Options for async commands"""
    function = click.option(
        "--detach",
        default=False,
        type=bool,
        help="Returns the created query ID to allow for asynchronous querying.",
    )(function)

    function = click.option(
        "-t",
        "--timeout",
        type=int,
        default=DEFAULT_QUERY_TIMEOUT,
        help="Sets the timeout to wait for query completion. Pass 0 to remove any timeout.",
    )(function)

    return function


def config_validation_options(function: Callable) -> Callable:
    """Common options when validating or committing configs"""
    function = click.option(
        "--config-dir",
        type=str,
        required=False,
        default=os.getcwd(),
        help="Path to directory containing Transform yaml models",
    )(function)
    return function


def query_options(function: Callable) -> Callable:
    """Common options for a query, adding support for limit, order, and cache-mode"""
    function = click.option(
        "--config-dir",
        type=str,
        required=False,
        help="Path to directory containing Transform yaml models to execute query against",
    )(function)

    function = click.option(
        "--cache-mode",
        type=click.Choice(
            list(map(lambda x: x.value, CacheMode)),
            case_sensitive=False,
        ),
        # Cast as CacheMode type
        callback=lambda ctx, param, value: CacheMode(value) if value is not None else None,
    )(function)

    function = click.option(
        "--order",
        required=False,
        multiple=True,
        help='Metrics or dimensions to order by ("-" in front of a column means descending). For example: --order -ds',
    )(function)

    function = click.option(
        "--limit",
        required=False,
        type=str,
        help="Limit the number of rows out(Default: 100) using an int or 'inf' for no limit. For example: --limit 100 or --limit inf",
    )(function)

    function = click.option(
        "--where",
        required=False,
        type=str,
        default=None,
        help="SQL-like where statement provided as a string. For example: --where \"ds = '2020-04-15'\"",
    )(function)

    function = click.option(
        "--time-constraint",
        required=False,
        type=str,
        default=None,
        help="Optional TimeConstraint on query (written as a WHERE clause, ie: `ds < '2020-01-01')",
    )(function)

    function = click.option(
        "--time-granularity",
        type=click.Choice(
            list(map(lambda x: x.value, TimeGranularity)),
            case_sensitive=False,
        ),
        # Cast as TimeGranularity type
        callback=lambda ctx, param, value: TimeGranularity(value) if value is not None else None,
    )(function)

    function = click.option(
        "--time-comparison",
        type=click.Choice(
            list(map(lambda x: x.value, PercentChange)),
            case_sensitive=False,
        ),
        # Cast as PercentChange type
        callback=lambda ctx, param, value: PercentChange(value) if value is not None else None,
    )(function)

    return function


def validate_query_args(limit: Optional[str]) -> None:
    """Logic to perform query validation before sending to the MQL Server"""
    if limit and not (limit.isnumeric() or limit in [INF, NEGATIVE_ONE]):
        click.echo(
            f"❌ limit must be an int (--limit {DEFAULT_LIMIT}) or {NEGATIVE_ONE} or {INF} (--limit {INF}) to specify no limit"
        )
        exit()


def metrics_and_dimensions_options(function: Callable) -> Callable:
    """TODO: could we validate that these exist using a callback?"""

    function = click.option(
        "--metrics",
        required=True,
        multiple=True,
        help="Metrics to query for: syntax is --metrics bookings. Submit a list of metrics by providing multiple --metrics flags.",
        callback=lambda ctx, param, value: __validate_and_normalize_query_inputs("--metrics ", value),
    )(function)
    function = click.option(
        "--dimensions",
        required=True,
        multiple=True,
        help="Dimensions to group by: syntax is --dimensions ds. Submit a list of metrics by providing multiple --dimensions flags.",
        callback=lambda ctx, param, value: __validate_and_normalize_query_inputs("--dimensions ", value),
    )(function)
    return function


def __validate_and_normalize_query_inputs(option_name: str, value: List[str]) -> List[str]:
    """A callback executed to normalize and validate query inputs for metrics and dimensions.

    TODO: Eventually we should validate that the metrics and dimensions actually exist on the org,
        avoiding a round trip and failed query. This would be a good place for that.
    """

    for v in value:
        # We have a comma-separated list of metrics or dimensions, explode this into a list and return
        if "," in v:
            return [i.strip() for i in v.split(",")]

    return list(value)


def no_active_mql_servers(org_name: str) -> None:
    """Read out if there are no mql servers found"""
    click.echo(f"‼️  The {org_name} organization doesn't currently have an active MQL Server.")
    click.echo(
        "An admin can create a new MQL Server at:\n\n    https://app.transformdata.io/settings/org/mql",
    )


def valid_datetime(dt_str: str) -> bool:
    """Returns true if string is valid iso8601 timestamp, false otherwise"""
    try:
        datetime.fromisoformat(dt_str)
    except Exception:
        try:
            datetime.fromisoformat(dt_str.replace("Z", "+00:00"))
        except Exception:
            return False
        return True
    return True


def enable_debug_log_file() -> None:  # noqa: D
    # Writing to the console would be ideal, but the newlines get messed up so it's hard to read.
    # e.g. log line looks like: Waiting for query2021-07-27 10:19:40,051 - http.client ...
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        filemode="a",
        filename="debug.log",
    )


def get_query_logs_url_for_id(query_id: str) -> str:
    """Returns the query log URL to the web UI given the query_id"""
    return f"https://app.transformdata.io/mql/query/{query_id}?tab=server_logs"


def exception_handler(func: Callable[..., Any]) -> Callable[..., Any]:  # type: ignore[misc]
    """Decorator to handle exceptions

    TODO: handle custom exceptions
    """

    @wraps(func)
    def wrapper(*args, **kwargs):  # type: ignore
        try:
            func(*args, **kwargs)
        except Exception as e:
            click.echo(f"\nERROR: {str(e)}")

    return wrapper



import time

from typing import Optional

from .models import MqlQueryStatusResp


class QueryRuntimeException(Exception):
    """Internal Exception type to surface query execution errors to the user"""

    query_id: str
    msg: Optional[str]
    start_time: Optional[float]

    def __init__(self, query_id: str, msg: Optional[str], start_time: Optional[float] = None) -> None:  # noqa: D
        self.msg = msg
        self.query_id = query_id
        self.start_time = start_time
        super().__init__(self.msg or "")

    def __str__(self) -> str:  # noqa: D
        if self.msg:
            msg = self.msg
        else:
            msg = "Failure 🧐 - query failed for unknown reason"
            if self.start_time:
                msg += f" after {time.time() - self.start_time:.2f} seconds..."
        return (
            msg
            + "\n💡 See MQL server logs for more details:\n"
            + f"\tmql stream-query-logs --query-id {self.query_id}\n"
        )

    @classmethod
    def from_query_response(cls, resp: MqlQueryStatusResp) -> "QueryRuntimeException":  # noqa: D
        if not resp.is_complete or resp.is_successful:
            raise Exception("Unable to build exception from unfailed query")
        return cls(query_id=resp.query_id, msg=resp.error)


class AuthException(Exception):
    """Exception to handle unauthorized access"""

    def __init__(self, msg: Optional[str] = None) -> None:  # noqa: D
        self.msg = msg
        super().__init__(self.msg or "")

    def __str__(self) -> str:  # noqa: D
        if self.msg:
            msg = self.msg
        else:
            msg = "Unable to find authentication"
        return (
            msg
            + "\n\tPlease retrieve an API key at https://app.transformdata.io/api_keys or run `mql setup` for multi-factor authentication"
        )


class URLException(Exception):
    """Invalid URL"""

    def __init__(self, url: str) -> None:  # noqa: D
        self.url = url

    def __str__(self) -> str:  # noqa: D
        return f"Detected invalid URL entry: {self.url}. Please try again."



import pandas as pd
import requests

from typing import Dict, List, Optional, Tuple

from .auth import TransformAuth
from .models import (
    CacheMode,
    Dimension,
    HealthReportStatus,
    Materialization,
    Metric,
    MqlMaterializeResp,
    MqlQueryStatusResp,
    MQLServer,
    MQLServerImage,
    ModelKey,
    PercentChange,
    Query,
    ServerHealthReport,
    TimeGranularity,
    UserState,
)
from .interfaces import MQLInterface


class MQLClient:
    """Wrapper that exposes MQLInterface and BackendInterface"""

    def __init__(self, api_key: Optional[str] = None, mql_server_url: Optional[str] = None) -> None:  # noqa: D
        self.context = TransformAuth(api_key, mql_server_url)

    def create_query(  # noqa: D
        self,
        metrics: List[str],
        dimensions: List[str],
        model_key_id: Optional[int] = None,
        where: Optional[str] = None,
        time_constraint: Optional[str] = None,
        time_granularity: Optional[str] = None,
        time_comparison: Optional[str] = None,
        order: Optional[List[str]] = None,
        limit: Optional[str] = None,
        cache_mode: Optional[str] = None,
        as_table: Optional[str] = None,
        allow_dynamic_cache: bool = True,
    ) -> MqlQueryStatusResp:
        """Builds a query in MQL and returns a MqlQueryStatusResp detailing the results."""
        model_key = self.get_model_key(model_key_id) if model_key_id else None
        query_id = self.context.mql_client.create_query(
            metrics,
            dimensions,
            model_key,
            where,
            time_constraint,
            TimeGranularity(time_granularity) if time_granularity else None,
            PercentChange(time_comparison) if time_comparison else None,
            order,
            limit,
            CacheMode(cache_mode) if cache_mode else None,
            as_table,
            allow_dynamic_cache,
        )
        return self.context.mql_client.get_query_status(query_id)

    def query(  # noqa: D
        self,
        metrics: List[str],
        dimensions: List[str],
        model_key_id: Optional[int] = None,
        where: Optional[str] = None,
        time_constraint: Optional[str] = None,
        time_granularity: Optional[str] = None,
        time_comparison: Optional[str] = None,
        order: Optional[List[str]] = None,
        limit: Optional[str] = None,
        cache_mode: Optional[str] = None,
        as_table: Optional[str] = None,
        allow_dynamic_cache: bool = True,
        timeout: Optional[int] = None,
    ) -> pd.DataFrame:
        """Builds a query in MQL and returns pandas dataframe of queried result.

        TODO: Now that MQL optionally returns synchronous results on query creation,
        the python interface should support opportunistically returning these rather
        than having to always poll for results
        """
        model_key = self.get_model_key(model_key_id) if model_key_id else None
        query_id = self.context.mql_client.create_query(
            metrics,
            dimensions,
            model_key,
            where,
            time_constraint,
            TimeGranularity(time_granularity) if time_granularity else None,
            PercentChange(time_comparison) if time_comparison else None,
            order,
            limit,
            CacheMode(cache_mode) if cache_mode else None,
            as_table,
            allow_dynamic_cache,
        )
        return self.get_query_dataframe(query_id, timeout)

    def list_queries(self, active_only: bool, limit: Optional[int] = None) -> List[Query]:
        """Retrieve a list of queries from MQL server."""
        return self.context.mql_client.get_queries(active_only=active_only, limit=limit)

    def list_metrics(self, model_key_id: Optional[int] = None) -> Dict[str, Metric]:
        """Returns a dictionary {metric_name: Metric}."""
        model_key = self.get_model_key(model_key_id) if model_key_id else None
        metrics = {m.name: m for m in self.context.mql_client.get_metrics(model_key)}
        metrics_metadata = self.context.backend_client.get_metrics_metadata()
        for data in metrics_metadata:
            metrics[data["name"]].input_metadata(data)
        return metrics

    def get_metric(self, metric_name: str, model_key_id: Optional[int] = None) -> Metric:
        """Retrieve a Metric Object by name."""
        model_key = self.get_model_key(model_key_id) if model_key_id else None
        metric = self.context.mql_client.get_metric_by_name(metric_name, model_key=model_key)
        metric.input_metadata(self.context.backend_client.get_metric_metadata_by_name(metric_name))
        return metric

    def create_materialization(  # noqa: D
        self,
        materialization_name: str,
        start_time: Optional[str],
        end_time: Optional[str],
        model_key_id: Optional[int] = None,
        output_table: Optional[str] = None,
        force: bool = False,
    ) -> MqlQueryStatusResp:
        """Implements Materialize (new) mutation in MQL/GQL and returns MqlQueryStatusResp object. Can be very expensive if a large time range is provided.

        Note: This is an asynchronous function (for synchronous function, see materialize)
        """
        model_key = self.get_model_key(model_key_id) if model_key_id else None
        query_id = self.context.mql_client.create_materialization(
            materialization_name, start_time, end_time, model_key, output_table, force
        )
        return self.context.mql_client.get_query_status(query_id)

    def materialize(  # noqa: D
        self,
        materialization_name: str,
        start_time: Optional[str],
        end_time: Optional[str],
        model_key_id: Optional[int] = None,
        output_table: Optional[str] = None,
        force: bool = False,
        timeout: Optional[int] = None,
    ) -> MqlMaterializeResp:
        """Builds materilization and returns MqlMaterializeResp object. Can be very expensive if a large time range is provided.

        Note: This is a synchronous function (for asynchronous function, see create_materialization)
        """
        model_key = self.get_model_key(model_key_id) if model_key_id else None
        query_id = self.context.mql_client.create_materialization(
            materialization_name, start_time, end_time, model_key, output_table, force
        )
        schema, table = self.get_materialization_result(query_id, timeout)
        return MqlMaterializeResp(table=table, schema=schema, query_id=query_id)

    def drop_materialization(  # noqa: D
        self,
        materialization_name: str,
        start_time: Optional[str],
        end_time: Optional[str],
        model_key_id: Optional[int] = None,
    ) -> MqlQueryStatusResp:
        """Implements drop materialization mutation in MQL/GQL. Can be very expensive if a large time range is provided.

        Note: This is an asynchronous function, the query may not be completed after the function returns.
        """
        model_key = self.get_model_key(model_key_id) if model_key_id else None
        query_id = self.context.mql_client.drop_materialization(materialization_name, start_time, end_time, model_key)
        return self.context.mql_client.get_query_status(query_id)

    def list_materializations(self, model_key_id: Optional[int] = None) -> List[Materialization]:
        """List the materializations for the Organization with their available metrics and dimensions."""
        model_key = self.get_model_key(model_key_id) if model_key_id else None
        return sorted(self.context.mql_client.get_materializations(model_key), key=lambda m: m.name)

    def commit_configs(self, config_dir: str, is_current: Optional[bool] = None) -> ModelKey:
        """Validates and commits yaml transform configs in the provided directory. Returns a ModelKey object."""
        return self.context.backend_client.commit_configs(config_dir, is_current)

    def validate_configs(self, config_dir: str) -> Tuple[str, str, str]:
        """Finds and validates yaml transform configs in the provided directory. Returns a tuple (repo, branch, commit)"""
        return self.context.backend_client.validate_configs(config_dir)

    def health_report(self) -> List[ServerHealthReport]:
        """Performs a health check on all mql servers within the organization and returns List[ServerHealthReport]."""
        servers = self.context.backend_client.get_org_mql_servers()
        # in the case where the user has a local override, we still want to run a health report
        if self.context.mql_client.rest_api_url not in [s.url for s in servers]:
            servers.append(
                MQLServer(id=0, name="Local Override", url=self.context.mql_client.rest_api_url, is_org_default=False)
            )
        result = []
        for s in servers:
            test_server = MQLInterface(self.context.auth_header, s.url)
            try:
                version, health = test_server.get_health_report()
                result.append(
                    ServerHealthReport(
                        name=s.name, status=HealthReportStatus.SUCCESS.value, url=s.url, version=version, servers=health
                    )
                )
            except Exception as e:  # noqa: D
                result.append(
                    ServerHealthReport(
                        name=s.name,
                        status=HealthReportStatus.FAIL.value,
                        url=s.url,
                        error_message="Unable to connect to MQL server",
                    )
                )
        return result

    def drop_cache(self) -> bool:
        """Drops the entire MQL cache. Only do this is the cache is somehow corrupt. Returns boolean about operation result."""
        return self.context.mql_client.drop_cache()

    def identify(self) -> UserState:
        """Returns a UserState object describing current user."""
        return self.context.user_state

    def ping(self) -> requests.Response:
        """Returns ping response from currently selected MQL server."""
        return self.context.mql_client.ping()

    def list_servers(self) -> List[MQLServer]:
        """Returns a list of MQL servers within the organization."""
        return self.context.backend_client.get_org_mql_servers()

    def poll_for_query_completion(self, query_id: str, timeout: Optional[int] = None) -> MqlQueryStatusResp:
        """Poll for query completion, displaying some progress indication."""
        return self.context.mql_client.poll_for_query_completion(query_id, timeout)

    def get_materialization_result(self, query_id: str, timeout: Optional[int] = None) -> Tuple[str, str]:
        """Retrieves materialized table resulting from successful materialization query."""
        return self.context.mql_client.get_materialization_result(query_id, timeout)

    def get_query_dataframe(self, query_id: str, timeout: Optional[int] = None) -> pd.DataFrame:
        """Retrieves query results in Pandas DataFrame format.

        Includes logic for:
        1. Polling for completion
        2. Paging through results
        3. Converting JSON to a Pandas DataFrame
        """
        return self.context.mql_client.get_query_dataframe(query_id, timeout)

    def get_model_key(self, model_key_id: int) -> ModelKey:
        """Retrieves ModelKey object given the id."""
        return self.context.backend_client.get_model_key(model_key_id)

    def latest_mql_image(self) -> MQLServerImage:
        """Retrieves the latest MQL server image."""
        return self.context.backend_client.get_latest_mql_server_image()

    def get_dimension_values(
        self,
        metric_name: str,
        dimension_name: str,
        model_key_id: Optional[int] = None,
        page_number: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[str]:
        """Retrieves a list of dimension values given a [metric_name, dimension_name]."""
        model_key = self.get_model_key(model_key_id) if model_key_id else None
        return self.context.mql_client.get_dimension_values(
            metric_name, dimension_name, model_key=model_key, page_number=page_number, page_size=page_size
        )

    def list_dimensions(self, model_key_id: Optional[int] = None) -> Dict[str, Dimension]:
        """Retrieves a list of all Dimension objects."""
        model_key = self.get_model_key(model_key_id) if model_key_id else None
        dimensions = self.context.mql_client.get_dimensions(model_key)
        return {dim.name: dim for dim in dimensions}



import json

import click
import time
import textwrap

import pandas as pd
from halo import Halo
from typing import Dict, List, Optional

from log_symbols import LogSymbols
from packaging.version import parse
from update_checker import UpdateChecker  # type: ignore

from .cli_context import CLIContext
from .cli_utils import (
    metrics_and_dimensions_options,
    query_options,
    config_validation_options,
    async_options,
    no_active_mql_servers,
    validate_query_args,
    valid_datetime,
    enable_debug_log_file,
    get_query_logs_url_for_id,
    query_observability_options,
    exception_handler,
)
from ..models import (
    CacheMode,
    PercentChange,
    Query,
    HealthReportStatus,
    DEFAULT_LIMIT,
    INF,
    NEGATIVE_ONE,
    TimeGranularity,
)
from .. import PACKAGE_NAME, __version__

pass_config = click.make_pass_decorator(CLIContext, ensure=True)

MAX_LIST_OBJECT_ELEMENTS = 5


@click.group()
@click.option("-v", "--verbose", is_flag=True)
@click.option("--debug-log-file", is_flag=True)
@pass_config
def cli(config: CLIContext, verbose: bool, debug_log_file: bool) -> None:  # noqa: D
    config.verbose = verbose

    if debug_log_file:
        enable_debug_log_file()

    checker = UpdateChecker()
    result = checker.check(PACKAGE_NAME, __version__)
    # result is None when an update was not found or a failure occured
    if result:
        # Note: As the CLI and API stabilize, we can be a little less aggressive about forcing upgrade
        # For now, we should do so to minimize the likelihood of out of date CLIs causing unexpected behavior
        click.secho(
            "Warning: A new version of the MQL CLI is available.",
            bold=True,
            fg="red",
        )

        click.echo(
            f"Please update to version {result.available_version}, released {result.release_date} by running:\n\t$ pip install --upgrade {PACKAGE_NAME}",
        )


@cli.command()
@pass_config
def contact(config: CLIContext) -> None:
    """Instructions for how to contact Transform for help."""
    click.echo(
        "🛎  We're here to help. Contact a friendly Transformer at support@transformdata.io or over a shared Slack if your org has one set up.",
    )


@cli.command()
@click.option(
    "-k",
    "--api-key",
    type=str,
    help="Your Transform API key used for authentication",
)
@click.option(
    "-o",
    "--mql-override",
    type=str,
    help="Override the regular MQL server URL",
    default=None,
)
@click.option(
    "-r",
    "--reset",
    type=bool,
    help="Reset the mql setup.",
    is_flag=True,
    default=False,
)
@click.option(
    "--remove-override",
    type=bool,
    help="Remove override on the regular MQL server URL",
    is_flag=True,
    default=False,
)
@click.option(
    "--unpin-model",
    type=bool,
    help="Remove pinned model ID and revert to the primary model for your organization",
    is_flag=True,
    default=False,
)
@pass_config
@exception_handler
def setup(
    config: CLIContext,
    api_key: Optional[str],
    mql_override: Optional[str],
    remove_override: bool,
    reset: bool,
    unpin_model: bool,
) -> None:
    """Guides user through CLI setup."""

    if api_key:
        config.api_key = api_key

    # Allow user to override their mql server via a flag
    if (remove_override or reset) and config.mql_server_url_config_override:
        config.mql_server_url_config_override = ""

    # Allow user to override their mql server via a flag
    if mql_override:
        config.mql_server_url_config_override = mql_override

    if unpin_model:
        config.unpin_model()

    if mql_override or remove_override or api_key:
        exit()

    # If the user doesn't yet have auth creds, wants to override existing cred, or token is expired
    if not config.just_authenticated and (
        not config.is_authenticated
        or click.confirm(
            f"We've found existing credentials for {config.user.user_name} within the {config.org.name} organization.\nWould you like to provide new credentials, thus resetting the existing credentials?"
        )
    ):
        config.prompt_authentication()

    # If the user has an existing MQL server override, let them know and ask if they want it removed
    if config.mql_server_url_config_override and click.confirm(
        config.mql_server_url_status + "\nWould you like to remove the MQL server URL override shown above?"
    ):
        config.mql_server_url_config_override = ""

    # Allow user to override their mql server via a flag
    if mql_override:
        config.mql_server_url_config_override = mql_override


@cli.command()
@pass_config
@exception_handler
def identify(config: CLIContext) -> None:
    """Identify the currently authenticated user."""
    config.identify()


@cli.command()
@pass_config
def version(config: CLIContext) -> None:
    """Print the current version of the MQL CLI."""
    click.echo(__version__)


@cli.command()
@pass_config
@exception_handler
@click.option("--force", "-f", is_flag=True, default=False)
def drop_cache(config: CLIContext, force: bool) -> None:
    """Drop the MQL cache. Only necessary if there is evidence cache corruption."""
    drop_confirm_msg = (
        click.style("WARNING", fg="yellow") + ": This can be an expensive operation.\nDo you want to drop the cache?"
    )
    if not force and not click.confirm(
        drop_confirm_msg,
        abort=False,
    ):
        click.echo("❌ Aborted dropping the MQL cache.")
        exit()

    spinner = Halo(text="Initiating drop-cache query... this can take a little while", spinner="dots")
    spinner.start()

    resp = config.mql.drop_cache()
    if not resp:
        raise click.ClickException("‼️ Failed to drop the cache.")
    spinner.succeed("💥 Successfully dropped MQL cache.")


@cli.command()
@pass_config
@exception_handler
def ping(config: CLIContext) -> None:
    """Perform basic HTTP health check against configured MQL server."""
    tic = time.perf_counter()
    resp = config.mql.ping()
    click.echo(
        f"🏓 Received HTTP {resp.status_code} code from MQL in {time.perf_counter() - tic:0.4f} seconds. \n {resp.text}"
    )


@cli.command()
@pass_config
@exception_handler
def list_servers(config: CLIContext) -> None:
    """Lists available MQL servers."""
    servers = config.mql.list_servers()
    if len(servers) == 0:
        no_active_mql_servers(config.org.name)
        exit()

    click.echo(f"🖨  We've found {len(servers)} MQL servers for the {config.org.name} organization. ⭐️ - Primary.")

    for s in servers:
        click.echo(f"• {'⭐️ ' if s.is_org_default else ''}{click.style(s.name, bold=True, fg='yellow')}: {s.url}")


@cli.command()
@pass_config
@exception_handler
def health_report(config: CLIContext) -> None:
    """Completes a health check on MQL servers."""
    spinner = Halo(text="Checking health of MQL servers...", spinner="dots")
    spinner.start()
    servers = config.mql.health_report()
    if len(servers) == 0:
        spinner.fail()
        no_active_mql_servers(config.org.name)
        exit()
    spinner.succeed("Successfully built health report!")

    click.echo(
        f"🏥 Health Report for {len(servers)} MQL Servers at {config.org.name}",
    )
    for s in servers:
        if s.status == HealthReportStatus.FAIL:
            click.echo(
                f"• ❌ {click.style(s.name, bold=True, fg=('red'))}:  Unable to connect to MQL server at url {s.url}."
            )
        else:
            click.echo(f"• {click.style(s.name, bold=True, fg='yellow')}: {s.url} running commit {s.version}")
            for h in s.servers:
                if h.status == HealthReportStatus.SUCCESS.value:
                    click.echo(f"  • ✅ {click.style(h.name, bold=True, fg=('green'))}: No Errors")
                else:
                    click.echo(f"  • ❌ {click.style(h.name, bold=True, fg=('red'))}:  {h.error_message}")


@cli.command()
@async_options
@metrics_and_dimensions_options
@query_options
@pass_config
@click.pass_context
@click.option(
    "--as-table",
    required=False,
    type=str,
)
@click.option(
    "--csv",
    type=click.File("wb"),
    required=False,
    help="Provide filepath for dataframe output to csv",
)
@click.option(
    "--explain",
    is_flag=True,
    required=False,
    default=False,
    help="In the query output, show the query that was executed against the data warehouse",
)
@click.option(
    "--decimals",
    required=False,
    default=2,
    help="Choose the number of decimal places to round for the numerical values",
)
@query_observability_options
@exception_handler
def query(
    ctx: click.core.Context,
    config: CLIContext,
    detach: bool,
    timeout: int,
    metrics: List[str],
    dimensions: List[str],
    where: Optional[str] = None,
    time_constraint: Optional[str] = None,
    time_granularity: Optional[TimeGranularity] = None,
    time_comparison: Optional[PercentChange] = None,
    order: Optional[List[str]] = None,
    limit: Optional[str] = None,
    cache_mode: Optional[CacheMode] = None,
    config_dir: Optional[str] = None,
    as_table: Optional[str] = None,
    csv: Optional[click.utils.LazyFile] = None,
    explain: bool = False,
    decimals: int = 2,
    web: bool = False,
    debug: bool = False,
    allow_dynamic_cache: bool = True,
) -> None:
    """Create a new MQL query, polls for completion and assembles a DataFrame from the response."""
    if not limit:
        limit = DEFAULT_LIMIT
    elif limit == INF:
        limit = NEGATIVE_ONE

    validate_query_args(limit)

    # Note: calling config.resolve_query_model_key may initiate a dialog if the user has a pinned model
    # Purposefully pulling this out before the spinner
    model_key = config.resolve_query_model_key(config_dir)

    start = time.time()
    spinner = Halo(text="Initiating query…", spinner="dots")
    spinner.start()
    query_id = config.mql.create_query(
        model_key_id=model_key.id if model_key else None,
        metrics=metrics,
        dimensions=dimensions,
        where=where,
        time_constraint=time_constraint,
        time_granularity=time_granularity.value if time_granularity else None,
        time_comparison=time_comparison.value if time_comparison else None,
        order=order,
        limit=limit,
        cache_mode=cache_mode.value if cache_mode else None,
        as_table=as_table,
        allow_dynamic_cache=allow_dynamic_cache,
    ).query_id

    spinner.succeed(f"Query initialized: {query_id}")
    if detach:
        return
    elif explain:
        sql = config.mql_client.explain_query_sql(query_id)
        if debug:
            ctx.invoke(stream_query_logs, query_id=query_id)
        click.echo(f"\n🔎 SQL executed for successful query (remove --explain to see data): \n{sql}")
        exit()

    spinner.start("Retrieving results")

    df: Optional[pd.DataFrame] = None
    try:
        df = config.mql.get_query_dataframe(query_id, timeout)
        final_message = f"Success 🦄 - query completed after {time.time() - start:.2f} seconds"
        final_symbol = LogSymbols.SUCCESS
    except Exception as e:
        final_message = str(e)
        final_symbol = LogSymbols.ERROR

    if debug:
        ctx.invoke(stream_query_logs, query_id=query_id)

    spinner.stop_and_persist(symbol=final_symbol.value, text=final_message)

    # Show the data if returned successfully
    if df is not None:
        if df.empty:
            click.echo("Successful MQL query returned an empty result set.")
        elif csv is not None:
            df.to_csv(csv, index=False)
            click.echo(f"Successfully written query output to {csv.name}")
        else:
            # NOTE: remove `to_string` if no pandas dependency is < 1.1.0
            if parse(pd.__version__) >= parse("1.1.0"):
                click.echo(df.to_markdown(index=False, floatfmt=f".{decimals}f"))
            else:
                click.echo(df.to_string(index=False, float_format=lambda x: format(x, f".{decimals}f")))

    if web:
        click.launch(get_query_logs_url_for_id(query_id))


@cli.command()
@click.option(
    "--materialization-name",
    required=True,
    type=str,
    help="Name of materialization to drop",
)
@click.option(
    "--start-time",
    required=False,
    type=str,
    help="iso8601 timestamp to drop from",
)
@click.option(
    "--end-time",
    required=False,
    type=str,
    help="iso8601 timestamp to drop to",
)
@async_options
@query_observability_options
@pass_config
@click.pass_context
@exception_handler
def drop_materialization(
    ctx: click.core.Context,
    config: CLIContext,
    detach: bool,
    timeout: int,
    materialization_name: str,
    start_time: Optional[str] = None,
    end_time: Optional[str] = None,
    config_dir: Optional[str] = None,
    web: bool = False,
    debug: bool = False,
) -> None:
    """***NEW*** Create a new MQL drop materialization query, polls for completion"""
    # Note: calling config.resolve_query_model_key may initiate a dialog if the user has a pinned model
    # Purposefully pulling this out before the spinner
    model_key = config.resolve_query_model_key(config_dir)

    if start_time and not valid_datetime(start_time):
        click.ClickException("start_time must be valid iso8601 timestamp")
    if end_time and not valid_datetime(end_time):
        click.ClickException("end_time must be valid iso8601 timestamp")

    if start_time is None:
        if not click.confirm(
            "You haven't provided a start_time. This means we will drop the materialization from the beginning of time."
            "This may be expensive. Are you sure you want to continue?"
        ):
            click.echo("Exiting")
            return

    start = time.time()
    spinner = Halo(text="Initiating drop materialization query…", spinner="dots")
    spinner.start()

    query_id = config.mql.drop_materialization(
        model_key_id=model_key.id if model_key else None,
        materialization_name=materialization_name,
        start_time=start_time,
        end_time=end_time,
    ).query_id

    spinner.succeed(f"Query initialized: {query_id}")
    if detach:
        return

    spinner.start("Waiting for query")
    try:
        config.mql.poll_for_query_completion(query_id, timeout)
        final_message = f"Success 🦄 - drop materialization query completed after {time.time() - start:.2f} seconds."
        final_symbol = LogSymbols.SUCCESS
    except Exception as e:
        final_message = str(e)
        final_symbol = LogSymbols.ERROR

    if debug:
        ctx.invoke(stream_query_logs, query_id=query_id)

    spinner.stop_and_persist(symbol=final_symbol.value, text=final_message)

    if web:
        click.launch(get_query_logs_url_for_id(query_id))


@cli.command()
@click.option(
    "--materialization-name",
    required=True,
    type=str,
    help="Name of materialization to materialize",
)
@click.option(
    "--start-time",
    required=False,
    type=str,
    help="iso8601 timestamp to materialize from",
)
@click.option(
    "--end-time",
    required=False,
    type=str,
    help="iso8601 timestamp to materialize to",
)
@click.option(
    "--output-table",
    required=False,
    type=str,
    help="Write materialized result to specified table of format '<schema>.<table_name>'",
)
@click.option("--force", "-f", is_flag=True, default=False)
@async_options
@query_observability_options
@pass_config
@click.pass_context
@exception_handler
def materialize(
    ctx: click.core.Context,
    config: CLIContext,
    detach: bool,
    timeout: int,
    materialization_name: str,
    start_time: Optional[str] = None,
    end_time: Optional[str] = None,
    config_dir: Optional[str] = None,
    output_table: Optional[str] = None,
    force: bool = False,
    web: bool = False,
    debug: bool = False,
) -> None:
    """Create a new MQL materialization query, polls for completion and returns materialized table id"""
    # Note: calling config.resolve_query_model_key may initiate a dialog if the user has a pinned model
    # Purposefully pulling this out before the spinner
    model_key = config.resolve_query_model_key(config_dir)

    if start_time and not valid_datetime(start_time):
        click.ClickException("start_time must be valid iso8601 timestamp")
    if end_time and not valid_datetime(end_time):
        click.ClickException("end_time must be valid iso8601 timestamp")

    if start_time is None:
        if not click.confirm(
            "You haven't provided a start_time. This means we will materialize from the beginning of time. This may be expensive. Are you sure you want to continue?"
        ):
            click.echo("Exiting")
            return

    start = time.time()
    spinner = Halo(text="Initiating materialization query…", spinner="dots")
    spinner.start()

    query_id = config.mql.create_materialization(
        model_key_id=model_key.id if model_key else None,
        materialization_name=materialization_name,
        start_time=start_time,
        end_time=end_time,
        output_table=output_table,
        force=force,
    ).query_id

    spinner.succeed(f"Query initialized: {query_id}")
    if detach:
        return

    spinner.start("Waiting for query")
    try:
        schema, table = config.mql.get_materialization_result(query_id, timeout)
        materialized_at = f"Materialized table: {schema}.{table}" if schema is not None else ""
        final_message = (
            f"Success 🦄 - materialize query completed after {time.time() - start:.2f} seconds." f"{materialized_at}"
        )
        final_symbol = LogSymbols.SUCCESS
    except Exception as e:
        final_message = str(e)
        final_symbol = LogSymbols.ERROR

    if debug:
        ctx.invoke(stream_query_logs, query_id=query_id)

    spinner.stop_and_persist(symbol=final_symbol.value, text=final_message)

    if web:
        click.launch(get_query_logs_url_for_id(query_id))


@cli.command()
@pass_config
@exception_handler
@click.option("--search", required=False, type=str, help="Filter available materializations by this search term")
def list_materializations(config: CLIContext, search: Optional[str] = None) -> None:
    """List the materializations for the Organization with their available metrics and dimensions."""

    spinner = Halo(text="Looking for all available materializations...", spinner="dots")
    spinner.start()

    model_key = config.resolve_query_model_key()
    materializations = config.mql.list_materializations(model_key.id if model_key else None)
    if not materializations:
        spinner.fail("List of materializations unavailable.")
        exit()

    filter_msg = ""
    if search is not None:
        count = len(materializations)
        materializations = [m for m in materializations if search.lower() in m.name.lower()]
        filter_msg = f" matching `{search}`, of a total of {count} available"

    spinner.succeed(
        f"🌱 We've found {len(materializations)} materializations for the {config.org.name} organization{filter_msg}."
    )
    click.echo(
        'The list below shows materializations in the format of "materialization: list of available metrics, then dimensions"'
    )
    for m in materializations:
        dimensions = sorted(m.dimensions)
        metrics = sorted(m.metrics)
        click.echo(
            f"• {click.style(m.name, bold=True, fg='green')}:"
            + (f"\nMetrics: {', '.join(metrics[:MAX_LIST_OBJECT_ELEMENTS])}")
            + (
                f" and {len(metrics) - MAX_LIST_OBJECT_ELEMENTS} more"
                if len(metrics) > MAX_LIST_OBJECT_ELEMENTS
                else ""
            )
            + (f"\nDimensions: {', '.join(dimensions[:MAX_LIST_OBJECT_ELEMENTS])}")
            + (
                f" and {len(dimensions) - MAX_LIST_OBJECT_ELEMENTS} more"
                if len(dimensions) > MAX_LIST_OBJECT_ELEMENTS
                else ""
            )
            + (f"\ndestination table: {m.destination_table or m.name}")
        )


@cli.command()
@pass_config
@click.option("--metric-name", required=True, type=str, help="Metric that is associated with the dimension")
@click.option("--dimension-name", required=True, type=str, help="Dimension to query")
@click.option("--page-number", required=False, type=int, hidden=True)
@click.option("--page-size", required=False, type=int, hidden=True)
@exception_handler
def get_dimension_values(
    config: CLIContext,
    dimension_name: str,
    metric_name: str,
    page_number: Optional[int] = None,
    page_size: Optional[int] = None,
) -> None:
    """List all dimension values that are queryable through materialized tables."""

    spinner = Halo(
        text=f"Retrieving dimension values for dimension {dimension_name} of metric {metric_name}...", spinner="dots"
    )
    spinner.start()

    model_key = config.resolve_query_model_key()
    try:
        dim_vals = sorted(
            config.mql.get_dimension_values(
                metric_name,
                dimension_name,
                model_key.id if model_key else None,
                page_number=page_number,
                page_size=page_size,
            )
        )
    except Exception as e:
        spinner.fail()
        click.echo(
            textwrap.dedent(
                f"""\
                ❌ Failed to query dimension values for dimension {dimension_name} of metric {metric_name}.
                💡 Ensure that this query can be done via pre-materialized results by creating a
                       materialization that includes the desired metric and dimension

                    ERROR: {str(e)}
                """
            )
        )
        exit(1)

    spinner.succeed(
        f"🌱 We've found {len(dim_vals)} dimension values for dimension {dimension_name} of metric {metric_name}."
    )
    for dim_val in dim_vals:
        click.echo(f"• {click.style(dim_val, bold=True, fg='green')}")


@cli.command()
@pass_config
@click.option("--search", required=False, type=str, help="Filter available metrics by this search term")
@click.option("--show-all-dims", is_flag=True, default=False, help="Show all dimensions associated with a metric.")
@exception_handler
def list_metrics(config: CLIContext, show_all_dims: bool = False, search: Optional[str] = None) -> None:
    """List the metrics for the Organization with their available dimensions.

    Automatically truncates long lists of dimensions, pass --show-all-dims to see all.
    """

    # Note: calling config.resolve_query_model_key may initiate a dialog if the user has a pinned model
    # Purposefully pulling this out before the spinner

    spinner = Halo(text="Looking for all available metrics...", spinner="dots")
    spinner.start()

    model_key = config.resolve_query_model_key()
    metrics = sorted(config.mql.list_metrics(model_key.id if model_key else None).values(), key=lambda m: m.name)
    if not metrics:
        spinner.fail("List of metrics unavailable.")

    filter_msg = ""
    if search is not None:
        num_org_metrics = len(metrics)
        metrics = [m for m in metrics if search.lower() in m.name.lower()]
        filter_msg = f" matching `{search}`, of a total of {num_org_metrics} available"

    spinner.succeed(f"🌱 We've found {len(metrics)} metrics for the {config.org.name} organization{filter_msg}.")
    click.echo('The list below shows metrics in the format of "metric_name: list of available dimensions"')
    num_dims_to_show = MAX_LIST_OBJECT_ELEMENTS
    for m in metrics:
        # sort dimensions by whether they're local first(if / then global else local) then the dim name
        dimensions = sorted(map(lambda d: d.name, filter(lambda d: "/" not in d.name, m.dimensions))) + sorted(
            map(lambda d: d.name, filter(lambda d: "/" in d.name, m.dimensions))
        )
        if show_all_dims:
            num_dims_to_show = len(dimensions)
        click.echo(
            f"• {click.style(m.name, bold=True, fg='green')}: {', '.join(dimensions[:num_dims_to_show])}"
            + (f" and {len(dimensions) - num_dims_to_show} more" if len(dimensions) > num_dims_to_show else "")
        )


@cli.command()
@pass_config
@exception_handler
def list_dimensions(config: CLIContext) -> None:
    """List all unique dimensions for the Organization."""

    spinner = Halo(text="Looking for all available dimensions...", spinner="dots")
    spinner.start()

    model_key = config.resolve_query_model_key()
    dimensions = sorted(config.mql.list_dimensions(model_key.id if model_key else None).values(), key=lambda d: d.name)
    if not dimensions:
        spinner.fail("List of dimensions unavailable.")

    spinner.succeed(f"🌱 We've found {len(dimensions)} unique dimensions for the {config.org.name} organization.")
    for d in dimensions:
        click.echo(f"• {click.style(d.name, bold=True, fg='green')}")


@cli.command()
@pass_config
@click.option("--query-id", required=True, type=str, help="Query ID to stream logs for")
@exception_handler
def stream_query_logs(
    config: CLIContext,
    query_id: str,
) -> None:
    """Retrieve queries from mql server"""
    line_number = 0
    query_status_resp = None
    while query_status_resp is None or not query_status_resp.is_complete:
        logs, line_count = config.mql_client.get_logs_by_line(query_id, line_number)
        line_number += line_count
        query_status_resp = config.mql_client.get_query_status(query_id)
        color_map: dict = {"INFO": "green", "ERROR": "red", "WARNING": "yellow"}
        if line_count > 0:
            lines: List[dict] = [json.loads(line) for line in logs.split("\n")[0:-1]]
            formatted_lines = [
                f'[{click.style(str(line.get("level")), bold=True, fg=color_map.get(line.get("level"), "green"))}]'
                f'[{line.get("asctime")}]: {line.get("message")}'
                for line in lines
            ]
            click.echo("\n".join(formatted_lines))
        time.sleep(1)

    if query_status_resp is not None and query_status_resp.error:
        click.echo("Error message: " + query_status_resp.error)


@cli.command()
@click.option("--active-only", type=bool, default=False, help="Return active queries only")
@click.option("--limit", required=False, type=int, help="Limit the number of queries retrieved: syntax is --limit 100")
@pass_config
@exception_handler
def list_queries(
    config: CLIContext,
    active_only: bool,
    limit: Optional[int] = None,
) -> None:
    """Retrieve queries from mql server"""

    active_query_str = "active" if active_only else "non-active"
    spinner = Halo(text=f"Looking for all {active_query_str} queries...", spinner="dots")
    spinner.start()

    queries = config.mql.list_queries(active_only=active_only, limit=limit)

    spinner.succeed(f"🌱 We've found {len(queries)} {active_query_str} queries. Grouped by branch-commit,")

    # group queries by branch-commit
    grouped_queries: Dict[str, List[Query]] = {}
    for q in queries:
        branch_commit = f"{q.branch}-{q.commit}"
        if branch_commit not in grouped_queries:
            grouped_queries[branch_commit] = []
        grouped_queries[branch_commit].append(q)

    for group in grouped_queries:
        click.echo(click.style(group, bold=True, fg=("green")))
        for q in grouped_queries[group]:
            click.echo(f"• query_id={q.id}, status={q.status.value}")


@cli.command()
@config_validation_options
@pass_config
@exception_handler
def validate_configs(config: CLIContext, config_dir: str) -> None:
    """Validate yaml configs found in specified config directory"""

    spinner = Halo(text=f"Validating configs in '{config_dir}'...", spinner="dots")
    spinner.start()
    repo, branch, commit = config.mql.validate_configs(config_dir)
    spinner.succeed(f"🎉 Successfully validated configs for commit ({commit}) in repo ({repo}), on branch ({branch})")


@cli.command()
@config_validation_options
@click.option("--pin", type=bool, required=False, help="Whether to pin the model or not")
@click.option("--force-primary", is_flag=True, default=None, help="Make this model current")
@pass_config
@click.pass_context
@exception_handler
def commit_configs(
    ctx: click.core.Context,
    config: CLIContext,
    config_dir: str,
    pin: Optional[bool] = None,
    force_primary: Optional[bool] = None,
) -> None:
    """Commit yaml configs found in specified config directory"""
    spinner = Halo(text="Parsing and uploading configs…", spinner="dots")
    spinner.start()
    model = config.mql.commit_configs(config_dir, force_primary)
    spinner.succeed(
        f"🦄 Successfully committed configs id {model.id} on branch {model.branch} with commit {model.commit}"
    )

    if pin is None:
        if click.confirm("📌 Would you like to pin this model commit for future MQL queries?"):
            pin = True

    if pin:
        ctx.invoke(pin_model, model_id=model.id)
    else:
        click.echo(
            textwrap.dedent(
                f"""\
                💡 Pin these configs in the future:

                    mql pin-model --model-id {model.id}
                """
            )
        )


@cli.command()
@click.option("--model-id", type=int, required=True, help="Model id to pin for local queries")
@pass_config
@exception_handler
def pin_model(config: CLIContext, model_id: int) -> None:
    """Pin a model id from configs that are already committed to the MQL Server"""
    try:
        config.pinned_model_id = model_id
    except Exception:
        click.echo(
            textwrap.dedent(
                f"""\
                ❌ Failed to pin model {model_id}. This may not be a valid model id.
                💡 Commit and pin new configs:

                    mql commit-configs --config-dir <path> --pin true

                """
            )
        )
        exit(1)

    click.echo(
        textwrap.dedent(
            f"""\
            ✅ Successfully pinned model {model_id}.
            💡 Unpin this model in the future:

                mql unpin-model

            """
        )
    )


@cli.command()
@pass_config
@exception_handler
def unpin_model(config: CLIContext) -> None:
    """Unpin a model id"""
    config.unpin_model()
    click.echo(
        textwrap.dedent(
            """\
                ✅ Successfully unpinned model and reverted to default configs.
                💡 Commit and pin new configs:

                    mql commit-configs --config-dir <path> --pin true

            """
        )
    )


@cli.command()
@pass_config
@exception_handler
def latest_mql_image(config: CLIContext) -> None:
    """Outputs the latest MQL server image details"""

    server_image = config.mql.latest_mql_image()
    click.echo(
        textwrap.dedent(
            f"""\
                ✅ Successfully retrieved the latest MQL server image,
                    Service Name: {server_image.service_name}
                    Version Hash: {server_image.version_hash}
                    Download URL: {server_image.download_url}

            """
        )
    )


if __name__ == "__main__":
    cli()



import os
import time
import git
import hashlib

from typing import Tuple

LOCAL = "local"


def is_git_repo(path: str = None) -> bool:
    """Check is a path is a git repo"""
    try:
        git.Repo(path, search_parent_directories=True)
        return True
    except git.exc.InvalidGitRepositoryError:
        return False
    return False


def get_git_info(path: str = None) -> Tuple[str, str, str]:
    """Attempts to retrieve git repo info"""
    repo = git.Repo(path, search_parent_directories=True)  # will throw an error if this is not a git repo

    # NOTE: we are taking the name of the directory as the name of the repo
    # TODO: better handling of detached state
    repo_name = os.path.split(os.path.realpath(os.path.dirname(repo.common_dir)))[-1]
    branch = repo.active_branch.name if not repo.head.is_detached else f"detached_{repo.head.object.hexsha}"
    commit = repo.head.object.hexsha

    return repo_name, branch, commit


def local_commit_info() -> Tuple[str, str, str]:
    """Create a pseudo-git setup for validation from a non-git repo"""
    hasher = hashlib.sha1()
    hasher.update(str(time.time()).encode("utf-8"))
    commit = str(hasher.hexdigest())
    return LOCAL, LOCAL, commit


def resolve_local_commit_info(config_dir: str) -> Tuple[str, str, str]:
    """Includes logic for uploading Transform models via the CLI."""
    if is_git_repo(config_dir):
        repo, branch, commit = get_git_info(config_dir)
    else:
        repo, branch, commit = local_commit_info()

    # to differentiate between CI-based commits and local CLI-based commits we edit the commit sha
    commit += f"-dirty-{int(time.time())}"
    return repo, branch, commit



from .mql import MQLClient

__version__ = "0.0.56"
PACKAGE_NAME = "transform"

# mql gets imported if user is already authenticated
mql = None
try:
    mql = MQLClient()
except Exception as e:  # noqa: D
    pass



import click
import time
import webbrowser
import textwrap
import requests

from datetime import datetime, timedelta
from halo import Halo
from typing import Optional, Dict
from dateutil import parser

from ..auth import TransformAuth
from ..constants import (
    API_KEY_CONFIG_KEY,
    BEARER_TOKEN_CONFIG_KEY,
    PINNED_MODEL_ID_CONFIG_KEY,
)
from ..models import ModelKey, UserState
from ..mql import MQLClient


class CLIContext(TransformAuth):
    """Context configs for CLI. Inherits base TransformAuth."""

    def __init__(self) -> None:  # noqa: D
        self.verbose = False
        self.__user_state: Optional[UserState] = None
        self.__just_authenticated = False
        super().__init__()
        self.mql = MQLClient()

    @property
    def api_key(self) -> Optional[str]:  # noqa: D
        return super().api_key

    @TransformAuth.api_key.setter
    def api_key(self, api_key: str) -> None:  # noqa: D
        """Set the users' API key and record to the config file."""
        TransformAuth.api_key.fset(self, api_key)  # type: ignore
        self.identify()
        self.__just_authenticated = True
        click.echo("✨ Success! We've set your Transform API key. We're ready to query.\n")

    def remove_api_key_and_prompt_browser_login(self) -> None:
        """Remove API key from config and prompt browser login. Used when MFA is newly required for an org."""
        if self.config._get_config_value(API_KEY_CONFIG_KEY) is not None:
            self.config._remove_config_value(API_KEY_CONFIG_KEY)
        self.prompt_auth0_login()

    @property
    def bearer_token(self) -> Optional[str]:
        """Get bearer token from config, if exists."""
        token = self.config._get_config_value(BEARER_TOKEN_CONFIG_KEY)
        if (
            token is not None
            and self.bearer_token_expires_at is not None
            and parser.parse(self.bearer_token_expires_at) <= datetime.now()
        ):
            click.echo("\n⌛ Authentication token has expired.")
            self.prompt_auth0_login()
            return self.bearer_token

        return token

    @TransformAuth.bearer_token.setter
    def bearer_token(self, bearer_token: str) -> None:  # noqa: D
        TransformAuth.bearer_token.fset(self, bearer_token)  # type: ignore
        self.identify()
        self.__just_authenticated = True
        if self.bearer_token_expires_at is not None:
            expires_in_hours = round((parser.parse(self.bearer_token_expires_at) - datetime.now()).seconds / (60 * 60))
            click.echo(
                f"✨ Success! We've set your Transform authorization credentials. Please note that these credentials will expire in {expires_in_hours} hours. We're ready to query.\n"
            )

    @property
    def is_authenticated(self) -> bool:  # noqa: D
        if self.api_key is not None or self.bearer_token is not None:
            try:
                self.user_state  # Will error if not authed with correct method
                return True
            except Exception:
                return False
        return False

    @property
    def just_authenticated(self) -> bool:
        """Check if the user authenticated on this request. Used for determining if auth override should be prompted."""
        return self.__just_authenticated

    def prompt_authentication(self) -> None:
        """Prompt the user to choose a method of authenticating."""
        click.echo(
            textwrap.dedent(
                """\
                \n
                Please enter one of the options below to authenticate.\n
                1. Enter API key. If you don't have an API key, you can visit https://app.transformdata.io/api_keys to create one.
                2. Use browser to verify device and log in.\n
                Note that if your organization enforces multi-factor authentication, you will be required to use browser login."""
            )
        )
        valid_choices = [1, 2]
        auth_choice = None
        while auth_choice not in valid_choices:
            auth_choice = click.prompt("""Enter option number here""", type=int)
            if auth_choice not in valid_choices:
                click.echo(f"\nInput must be one of {valid_choices}. Please try again.")

        if auth_choice == 1:
            self.api_key = click.prompt("Enter API key", type=str, hide_input=True)
        elif auth_choice == 2:
            self.prompt_auth0_login()

    def prompt_auth0_login(self) -> None:
        """Prompt user to go through device authorization flow via Auth0.

        Documentation: https://auth0.com/docs/authorization/flows/call-your-api-using-the-device-authorization-flow
        """
        # Request device authorization code
        AUTH0_CLIENT_ID_CLI_PROD = "vaky0v5TpN5KiGq1kcxYh49CK0P6Akzg"
        AUTH0_DOMAIN_PROD = "https://tfd-main-prod.us.auth0.com"
        BACKEND_API_URL = "https://api.transformdata.io/test/v1/"
        device_code_response = requests.post(
            f"{AUTH0_DOMAIN_PROD}/oauth/device/code",
            headers={"content-type": "application/x-www-form-urlencoded"},
            data={"client_id": AUTH0_CLIENT_ID_CLI_PROD, "audience": BACKEND_API_URL, "scope": "offline_access"},
        ).json()

        expected_keys = ["verification_uri_complete", "user_code", "device_code", "interval", "expires_in"]
        if not all(key in device_code_response for key in expected_keys):
            raise click.ClickException(
                f"{device_code_response.get('error')}: {device_code_response.get('error_description')}"
            )

        device_verification_url = device_code_response["verification_uri_complete"]
        user_code = device_code_response["user_code"]
        device_code = device_code_response["device_code"]
        polling_interval_seconds = device_code_response["interval"]
        device_code_expires_at = datetime.now() + timedelta(seconds=int(device_code_response["expires_in"]))

        # Prompt user to log in via browser
        click.echo(
            textwrap.dedent(
                f"""\
                \n
                You should be directed to the login page in your browser automatically.
                If you don't see the login page, click this link: {device_verification_url}
                You will be asked to confirm this device code: {user_code}
                """
            )
        )
        webbrowser.open(device_verification_url)

        # Poll for authentication at recommended interval until you get a useful response or device code expires
        token_response: Dict[str, str] = {}
        token_response_error: Optional[str] = None
        device_code_expired = False
        while (token_response == {} or token_response_error == "authorization_pending") and not device_code_expired:
            time.sleep(polling_interval_seconds)
            token_response = requests.post(
                f"{AUTH0_DOMAIN_PROD}/oauth/token",
                headers={"content-type": "application/x-www-form-urlencoded"},
                data={
                    "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                    "device_code": device_code,
                    "client_id": AUTH0_CLIENT_ID_CLI_PROD,
                },
            ).json()
            token_response_error = token_response.get("error")
            device_code_expired = (
                token_response_error in ["expired_token", "invalid_grant"] or datetime.now() >= device_code_expires_at
            )

        # Store token if successful, else show error
        if "access_token" in token_response:
            self.bearer_token_expires_at = str(datetime.now() + timedelta(seconds=int(token_response["expires_in"])))
            self.bearer_token = token_response["access_token"]
        elif token_response_error is not None:
            raise click.ClickException(textwrap.dedent(token_response["error_description"]))
        else:
            raise click.ClickException("Authentication failed. Please try again.")
            self.prompt_auth0_login()

    @property
    def user_state(self) -> UserState:  # noqa: D
        if self.__user_state is None:
            try:
                self.__user_state = self.mql.identify()
            except Exception as e:
                # TODO: Figure out a better way to determine user's org is blocking API key use without doing hacky string matching!
                error_msg = str(e)
                mfa_msg = "Your organization requires multi-factor authentication. Please use browser login instead of API key for authentication."
                if mfa_msg in error_msg:
                    click.echo(mfa_msg)
                    self.remove_api_key_and_prompt_browser_login()
                    return self.user_state
                else:
                    raise click.ClickException(error_msg)
        return self.__user_state

    @property
    def pinned_model_id(self) -> Optional[int]:
        """Config override for the org's current model key. Primarily used for development"""
        out = self.config._get_config_value(PINNED_MODEL_ID_CONFIG_KEY)
        return int(out) if out else None

    @pinned_model_id.setter
    def pinned_model_id(self, model_id: int) -> None:  # noqa: D
        if not model_id:
            self.config._remove_config_value(PINNED_MODEL_ID_CONFIG_KEY)

        # fetch model row to ensure it's valid
        self.mql.get_model_key(model_id)
        self.config._set_config_value(PINNED_MODEL_ID_CONFIG_KEY, str(model_id))

    def unpin_model(self) -> None:  # noqa: D
        if self.pinned_model_id:
            self.config._remove_config_value(PINNED_MODEL_ID_CONFIG_KEY)

    @property
    def current_model(self) -> Optional[ModelKey]:  # noqa: D
        return self.user_state.current_model

    def resolve_query_model_key(self, config_dir: Optional[str] = None) -> Optional[ModelKey]:
        """Resolve which model key to use for a given query or materialization.

        * If the user has specified they want to read from a local config directory, we commit that
        * If the user has a model id pinned, we let them know in the logs.
        * Otherwise, we default to None so that the query server can resolve the proper model key for the org.
        """
        if config_dir:
            spinner = Halo(text="Parsing and uploading local configs…", spinner="dots")
            spinner.start()
            model_key = self.mql.commit_configs(config_dir)
            spinner.succeed("🌱 Successfully parsed local configs")
            return model_key

        if self.pinned_model_id:

            pinned_model = self.mql.get_model_key(self.pinned_model_id)
            msg = textwrap.dedent(
                f"""\
                    📌 We've found a locally pinned model that we will be using to complete this request, uploaded {pinned_model.created_at}.
                    """
            )

            # Only ask the user if they want to continue with the override if the pinned model is not the primary for the org
            if pinned_model.is_current:
                click.echo(
                    msg
                    + "\nThe current model on your MQL Server matches matched the locally pinned model and will be used in this request"
                )
                return None
            else:

                if self.current_model is None:
                    return None
                else:
                    click.echo(
                        msg
                        + textwrap.dedent(
                            f"""
                            This model does not match the current primary model for {self.org.name}, uploaded {self.current_model.created_at}.
                            To unpin this model for future queries: `mql unpin-model`
                            """
                        ).strip()
                    )
                    return pinned_model
        else:
            return None

    def identify(self) -> None:  # noqa: D
        msg = f"You are authenticated as {self.user_state.user.user_name} within the {self.user_state.organization.name} organization"

        if self.user_state.current_model:
            msg = (
                msg
                + f" using model at {self.user_state.current_model.repository}/{self.user_state.current_model.branch}:{self.user_state.current_model.commit}\n"
                + self.mql_server_url_status
            )
        else:
            msg = (
                msg
                + ". There are currently no metrics committed for your org, please see https://app.transformdata.io/install "
                + "for instructions on how to commit your first Transform metrics."
            )

        click.echo(msg)

    @property
    def mql_server_url_config_override(self) -> Optional[str]:  # noqa: D
        return super().mql_server_url_config_override

    @TransformAuth.mql_server_url_config_override.setter
    def mql_server_url_config_override(self, mql_server_url: str) -> None:  # noqa: D
        TransformAuth.mql_server_url_config_override.fset(self, mql_server_url)  # type: ignore
        if mql_server_url:
            click.echo(f"Success! We've overriden your MQL server to {mql_server_url}")



import re
from setuptools import find_packages, setup

# Read the version and package name from cli/__init__.py
# From https://gehrcke.de/2014/02/distributing-a-python-command-line-application/
# Your setup.py should not import your package for reading the version number.
# Instead, always read it directly. In this case, I used regular expressions for extracting it.
version = re.search('^__version__\\s*=\\s*"(.*)"', open("transform/__init__.py").read(), re.M)
PACKAGE_NAME = re.search('^PACKAGE_NAME\\s*=\\s*"(.*)"', open("transform/__init__.py").read(), re.M)
if not version or not PACKAGE_NAME:
    raise Exception("Failed to parse transform/__init__.py for version and package name.")

setup(
    name=PACKAGE_NAME.group(1),
    version=version.group(1),
    # Note, we could possibly lower this but haven't yet tested older versions and haven't yet been asked to :)
    python_requires=">=3.6",
    author="Transform Data",
    author_email="marco@transformdata.io",
    description=("Transform MQL Client Library"),
    long_description="",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    entry_points={"console_scripts": ["mql = transform.cli.mql_cli:cli"]},
    py_modules=["mql_cli"],
    # Package details
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    # Dependencies
    setup_requires=["setupmeta"],
    install_requires=[
        "gql",
        "gitpython",
        "requests>=2.23.0",
        "pytest",
        "click",
        "pyyaml",
        "update_checker",
        "tabulate",
        "pandas>=0.24.2,<2.0",
        "dataclasses",
        "halo",
        "validators",
        "transform_tools>=0.0.20",
        "arrow",
        "log_symbols>=0.0.14",
    ],
)
